from moto.core.responses import ActionResult, BaseResponse, EmptyResult
from moto.core.serialize import never_return

from .models import MediaPackagev2Backend, mediapackagev2_backends


class mediapackagev2Response(BaseResponse):
    """Handler for mediapackagev2 requests and responses."""

    RESPONSE_KEY_PATH_TO_TRANSFORMER = {
        # InputType is only returned on GET
        "CreateChannel.CreateChannelResponse.InputType": never_return
    }

    def __init__(self) -> None:
        super().__init__(service_name="mediapackagev2")

    @property
    def mediapackagev2_backend(self) -> MediaPackagev2Backend:
        return mediapackagev2_backends[self.current_account][self.region]

    def create_channel_group(self) -> ActionResult:
        channel_group_name = self._get_param("ChannelGroupName")
        group = self.mediapackagev2_backend.create_channel_group(
            channel_group_name=channel_group_name,
        )
        return ActionResult(result=group)

    def create_channel(self) -> ActionResult:
        channel_group_name = self._get_param("ChannelGroupName")
        channel_name = self._get_param("ChannelName")
        channel = self.mediapackagev2_backend.create_channel(
            channel_group_name=channel_group_name,
            channel_name=channel_name,
        )
        return ActionResult(result=channel)

    def get_channel_group(self) -> ActionResult:
        channel_group_name = self._get_param("ChannelGroupName")
        group = self.mediapackagev2_backend.get_channel_group(
            channel_group_name=channel_group_name,
        )
        return ActionResult(result=group)

    def list_channel_groups(self) -> ActionResult:
        max_results = self._get_int_param("maxResults")
        next_token = self._get_param("nextToken")

        groups, next_token = self.mediapackagev2_backend.list_channel_groups(
            max_results=max_results,
            next_token=next_token,
        )
        return ActionResult(result={"Items": groups, "NextToken": next_token})

    def get_channel(self) -> ActionResult:
        channel_group_name = self._get_param("ChannelGroupName")
        channel_name = self._get_param("ChannelName")
        channel = self.mediapackagev2_backend.get_channel(
            channel_group_name=channel_group_name,
            channel_name=channel_name,
        )
        return ActionResult(result=channel)

    def delete_channel_group(self) -> ActionResult:
        channel_group_name = self._get_param("ChannelGroupName")
        self.mediapackagev2_backend.delete_channel_group(
            channel_group_name=channel_group_name
        )
        return EmptyResult()

    def delete_channel(self) -> ActionResult:
        channel_group_name = self._get_param("ChannelGroupName")
        channel_name = self._get_param("ChannelName")
        self.mediapackagev2_backend.delete_channel(
            channel_group_name=channel_group_name,
            channel_name=channel_name,
        )
        return EmptyResult()



"""s3control base URL and path."""

from .responses import S3ControlResponse

url_bases = [
    r"https?://([0-9]+)\.s3-control\.(.+)\.amazonaws\.com",
]


url_paths = {
    r"{0}/v20180820/accesspoint$": S3ControlResponse.dispatch,
    r"{0}/v20180820/configuration/publicAccessBlock$": S3ControlResponse.dispatch,
    r"{0}/v20180820/accesspoint/(?P<name>[\w_:%-]+)$": S3ControlResponse.dispatch,
    r"{0}/v20180820/accesspoint/(?P<name>[\w_:%-]+)/policy$": S3ControlResponse.dispatch,
    r"{0}/v20180820/accesspoint/(?P<name>[\w_:%-]+)/policyStatus$": S3ControlResponse.dispatch,
    r"{0}/v20180820/async-requests/mrap.*$": S3ControlResponse.dispatch,
    r"{0}/v20180820/mrap/instances.*$": S3ControlResponse.dispatch,
    "{0}/v20180820/storagelens/(?P<storagelensid>[^/]+)$": S3ControlResponse.dispatch,
    "{0}/v20180820/storagelens/(?P<storagelensid>[^/]+)/tagging$": S3ControlResponse.dispatch,
    "{0}/v20180820/storagelens$": S3ControlResponse.dispatch,
    r"{0}/v20180820/tags/(?P<arn>[\w_:%-]+)$": S3ControlResponse.dispatch,
    "/v20180820/accesspoint": S3ControlResponse.dispatch,
    "/v20180820/configuration/publicAccessBlock": S3ControlResponse.dispatch,
    "/v20180820/accesspoint/<name>": S3ControlResponse.dispatch,
    "/v20180820/accesspoint/<name>/policy": S3ControlResponse.dispatch,
    "/v20180820/accesspoint/<name>/policyStatus": S3ControlResponse.dispatch,
    "/v20180820/async-requests/mrap/create": S3ControlResponse.dispatch,
    "/v20180820/async-requests/mrap/delete": S3ControlResponse.dispatch,
    "/v20180820/async-requests/mrap/put-policy": S3ControlResponse.dispatch,
    "/v20180820/async-requests/mrap/<path:request_token_arn>": S3ControlResponse.dispatch,
    "/v20180820/mrap/instances": S3ControlResponse.dispatch,
    "/v20180820/mrap/instances/<mrap_name>": S3ControlResponse.dispatch,
    "/v20180820/mrap/instances/<mrap_name>/policy": S3ControlResponse.dispatch,
    "/v20180820/mrap/instances/<mrap_name>/policystatus": S3ControlResponse.dispatch,
    "/v20180820/storagelens/<storagelensid>": S3ControlResponse.dispatch,
    "/v20180820/storagelens/<storagelensid>/tagging": S3ControlResponse.dispatch,
    "/v20180820/storagelens": S3ControlResponse.dispatch,
    "/v20180820/tags/<arn>": S3ControlResponse.dispatch,
}



from datetime import datetime
from typing import Any, SupportsFloat


def parse_expression(
    expression: str, results: list[dict[str, Any]]
) -> tuple[list[SupportsFloat], list[datetime]]:
    values: list[SupportsFloat] = []
    timestamps: list[datetime] = []
    for result in results:
        if result.get("id") == expression:
            values.extend(result["values"])
            timestamps.extend(result["timestamps"])
    return values, timestamps



import json
from urllib.parse import unquote

from moto.core.responses import BaseResponse

from .models import VPCLatticeBackend, vpclattice_backends


class VPCLatticeResponse(BaseResponse):
    def __init__(self) -> None:
        super().__init__(service_name="vpc-lattice")

    @property
    def backend(self) -> VPCLatticeBackend:
        return vpclattice_backends[self.current_account][self.region]

    def create_service(self) -> str:
        service = self.backend.create_service(
            auth_type=self._get_param("authType"),
            certificate_arn=self._get_param("certificateArn"),
            client_token=self._get_param("clientToken"),
            custom_domain_name=self._get_param("customDomainName"),
            name=self._get_param("name"),
            tags=self._get_param("tags"),
        )
        return json.dumps(service.to_dict())

    def get_service(self) -> str:
        path = unquote(self.path)
        service = self.backend.get_service(service_identifier=path.split("/")[-1])
        return json.dumps(service.to_dict())

    def list_services(self) -> str:
        max_results = self._get_param("MaxResults")
        next_token = self._get_param("NextToken")
        services, next_token = self.backend.list_services(
            max_results=max_results, next_token=next_token
        )
        response = {
            "items": [service.to_dict() for service in services],
            "nextToken": next_token,
        }
        return json.dumps(response)

    def create_service_network(self) -> str:
        sn = self.backend.create_service_network(
            auth_type=self._get_param("authType"),
            client_token=self._get_param("clientToken"),
            name=self._get_param("name"),
            sharing_config=self._get_param("sharingConfig"),
            tags=self._get_param("tags"),
        )
        return json.dumps(sn.to_dict())

    def get_service_network(self) -> str:
        path = unquote(self.path)
        service_network = self.backend.get_service_network(
            service_network_identifier=path.split("/")[-1]
        )
        return json.dumps(service_network.to_dict())

    def list_service_networks(self) -> str:
        max_results = self._get_param("MaxResults")
        next_token = self._get_param("NextToken")
        service_networks, next_token = self.backend.list_service_networks(
            max_results=max_results, next_token=next_token
        )
        response = {
            "items": [
                service_network.to_dict() for service_network in service_networks
            ],
            "nextToken": next_token,
        }
        return json.dumps(response)

    def create_service_network_vpc_association(self) -> str:
        assoc = self.backend.create_service_network_vpc_association(
            client_token=self._get_param("clientToken"),
            security_group_ids=self._get_param("securityGroupIds"),
            service_network_identifier=self._get_param("serviceNetworkIdentifier"),
            tags=self._get_param("tags"),
            vpc_identifier=self._get_param("vpcIdentifier"),
        )
        return json.dumps(assoc.to_dict())

    def create_rule(self) -> str:
        rule = self.backend.create_rule(
            action=self._get_param("action"),
            client_token=self._get_param("clientToken"),
            listener_identifier=self._get_param("listenerIdentifier"),
            match=self._get_param("match"),
            name=self._get_param("name"),
            priority=self._get_param("priority"),
            service_identifier=self._get_param("serviceIdentifier"),
            tags=self._get_param("tags"),
        )
        return json.dumps(rule.to_dict())

    def list_tags_for_resource(self) -> str:
        resource_arn = unquote(self._get_param("resourceArn"))
        tags = self.backend.list_tags_for_resource(resource_arn)
        return json.dumps({"tags": tags})

    def tag_resource(self) -> str:
        resource_arn = unquote(self._get_param("resourceArn"))
        tags = self._get_param("tags")
        self.backend.tag_resource(resource_arn, tags)
        return json.dumps({})

    def untag_resource(self) -> str:
        resource_arn = unquote(self._get_param("resourceArn"))
        tag_keys = self._get_param("tagKeys")
        self.backend.untag_resource(resource_arn=resource_arn, tag_keys=tag_keys)
        return json.dumps({})

    def create_access_log_subscription(self) -> str:
        sub = self.backend.create_access_log_subscription(
            resourceIdentifier=self._get_param("resourceIdentifier"),
            destinationArn=self._get_param("destinationArn"),
            client_token=self._get_param("clientToken"),
            serviceNetworkLogType=self._get_param("serviceNetworkLogType"),
            tags=self._get_param("tags"),
        )

        return json.dumps(sub.to_dict())

    def get_access_log_subscription(self) -> str:
        path = unquote(self.path)

        sub = self.backend.get_access_log_subscription(
            accessLogSubscriptionIdentifier=path.split("/")[-1]
        )

        return json.dumps(sub.to_dict())

    def list_access_log_subscriptions(self) -> str:
        subs = self.backend.list_access_log_subscriptions(
            resourceIdentifier=self._get_param("resourceIdentifier"),
            maxResults=self._get_int_param("maxResults"),
            nextToken=self._get_param("nextToken"),
        )

        return json.dumps({"items": [s.to_dict() for s in subs], "nextToken": ""})

    def update_access_log_subscription(self) -> str:
        path = unquote(self.path)

        sub = self.backend.update_access_log_subscription(
            accessLogSubscriptionIdentifier=path.split("/")[-1],
            destinationArn=self._get_param("destinationArn"),
        )

        return json.dumps(sub.to_dict())

    def delete_access_log_subscription(self) -> str:
        path = unquote(self.path)

        self.backend.delete_access_log_subscription(
            accessLogSubscriptionIdentifier=path.split("/")[-1]
        )

        return json.dumps({})

    def put_auth_policy(self) -> str:
        resourceId = unquote(self._get_param("resourceIdentifier"))
        policy = self._get_param("policy")

        auth_policy = self.backend.put_auth_policy(
            resourceIdentifier=resourceId,
            policy=policy,
        )

        response = {
            "policy": auth_policy.policy,
            "state": auth_policy.state,
        }
        return json.dumps(response)

    def get_auth_policy(self) -> str:
        resourceId = unquote(self._get_param("resourceIdentifier"))
        auth_policy = self.backend.get_auth_policy(resourceIdentifier=resourceId)

        response = {
            "policy": auth_policy.policy,
            "state": auth_policy.state,
            "createdAt": auth_policy.created_at,
            "lastUpdatedAt": auth_policy.last_updated_at,
        }
        return json.dumps(response)

    def delete_auth_policy(self) -> str:
        resourceId = unquote(self._get_param("resourceIdentifier"))

        self.backend.delete_auth_policy(resourceIdentifier=resourceId)

        return "{}"

    def put_resource_policy(self) -> str:
        resource_arn = unquote(self._get_param("resourceArn"))
        policy = self._get_param("policy")
        self.backend.put_resource_policy(
            resourceArn=resource_arn,
            policy=policy,
        )

        return "{}"

    def get_resource_policy(self) -> str:
        resource_arn = unquote(self._get_param("resourceArn"))

        resource_policy = self.backend.get_resource_policy(resourceArn=resource_arn)

        return json.dumps({"policy": resource_policy})

    def delete_resource_policy(self) -> str:
        resource_arn = unquote(self._get_param("resourceArn"))

        self.backend.delete_resource_policy(resourceArn=resource_arn)

        return "{}"



from __future__ import annotations

import itertools
import random
from typing import Any, cast


class ReservedInstancesOffering(dict[str, Any]):
    """Lightweight container for a Reserved Instance Offering.

    This is a simplified representation that captures commonly-used fields
    for testing without modeling full AWS behavior.
    """

    _filter_attributes = {
        "availability-zone": ["AvailabilityZone"],
        "instance-type": ["InstanceType"],
        "product-description": ["ProductDescription"],
        "offering-type": ["OfferingType"],
    }

    def __init__(self, **kwargs: Any):
        super().__init__(**kwargs)

    def get_filter_value(self, filter_name: str) -> Any:
        path = self._filter_attributes.get(filter_name)
        value: Any = self
        if not path:
            # Unknown filter -> don't match
            return None
        for key in path:
            value = (value or {}).get(key)
        return value


class ReservedInstancesBackend:
    """Backend mixin to support DescribeReservedInstancesOfferings.

    We generate a small, static set of offerings per region/zone to make
    local tests useful while keeping scope limited.
    """

    _ri_offerings_generated: bool = False
    _ri_offerings: dict[str, list[ReservedInstancesOffering]]

    def _ensure_reserved_offerings(self) -> None:
        if getattr(self, "_ri_offerings_generated", False):
            return

        # Build a handful of offerings for the current region based on its AZs
        region = self.region_name  # type: ignore[attr-defined]
        zones = [z.name for z in self.describe_availability_zones()]  # type: ignore[attr-defined]
        instance_types = [
            "t2.micro",
            "t3.small",
            "m5.large",
        ]
        offering_types = ["No Upfront", "Partial Upfront", "All Upfront"]
        product = "Linux/UNIX"

        offerings: list[ReservedInstancesOffering] = []
        rid_counter = 1
        for itype, zone, otype in itertools.product(
            instance_types, zones, offering_types
        ):
            rid = f"{region}-ri-off-{rid_counter:04d}"
            rid_counter += 1
            duration = 31536000  # 1 year in seconds
            fixed = 0.0 if otype == "No Upfront" else (float(random.randint(50, 300)))
            usage = 0.0 if otype == "All Upfront" else 0.05

            offerings.append(
                ReservedInstancesOffering(
                    ReservedInstancesOfferingId=rid,
                    InstanceType=itype,
                    AvailabilityZone=zone,
                    ProductDescription=product,
                    Duration=duration,
                    UsagePrice=usage,
                    FixedPrice=fixed,
                    CurrencyCode="USD",
                    InstanceTenancy="default",
                    OfferingClass="standard",
                    OfferingType=otype,
                    Marketplace=False,
                )
            )

        # Also add region-wide offerings (no specific AZ) for convenience
        for itype, otype in itertools.product(instance_types, offering_types):
            rid = f"{region}-ri-off-{rid_counter:04d}"
            rid_counter += 1
            offerings.append(
                ReservedInstancesOffering(
                    ReservedInstancesOfferingId=rid,
                    InstanceType=itype,
                    AvailabilityZone=None,
                    ProductDescription=product,
                    Duration=31536000,
                    UsagePrice=0.05,
                    FixedPrice=0.0 if otype == "No Upfront" else 100.0,
                    CurrencyCode="USD",
                    InstanceTenancy="default",
                    OfferingClass="standard",
                    OfferingType=otype,
                    Marketplace=False,
                )
            )

        if not hasattr(self, "_ri_offerings"):
            self._ri_offerings = {}
        self._ri_offerings[region] = offerings
        self._ri_offerings_generated = True

    def describe_reserved_instances_offerings(
        self, filters: dict[str, list[str]] | None = None
    ) -> list[dict[str, Any]]:
        self._ensure_reserved_offerings()
        region = self.region_name  # type: ignore[attr-defined]
        offerings = self._ri_offerings.get(region, [])

        if not filters:
            return cast(list[dict[str, Any]], offerings)

        # Simple filter predicate across supported keys
        def predicate(obj: ReservedInstancesOffering, ff: dict[str, list[str]]) -> bool:
            for name, values in ff.items():
                val = obj.get_filter_value(name)
                if val is None:
                    # Unsupported filter -> treat as non-match
                    return False
                if isinstance(val, list):
                    if not any(v in val for v in values):
                        return False
                else:
                    if val not in values:
                        return False
            return True

        return cast(
            list[dict[str, Any]], [o for o in offerings if predicate(o, filters)]
        )



"""Exceptions raised by the sdb service."""

from moto.core.exceptions import ServiceException


class SDBError(ServiceException):
    box_usage = 0.0055590278


class InvalidParameterError(SDBError):
    code = "InvalidParameterValue"


class InvalidDomainName(InvalidParameterError):
    def __init__(self, domain_name: str):
        super().__init__(f"Value ({domain_name}) for parameter DomainName is invalid. ")


class UnknownDomainName(SDBError):
    code = "NoSuchDomain"
    message = "The specified domain does not exist."



from moto.core.responses import ActionResult

from ._base_response import EC2BaseResponse


class IamInstanceProfiles(EC2BaseResponse):
    def associate_iam_instance_profile(self) -> ActionResult:
        instance_id = self._get_param("InstanceId")
        iam_instance_profile_name = self._get_param("IamInstanceProfile.Name")
        iam_instance_profile_arn = self._get_param("IamInstanceProfile.Arn")
        iam_association = self.ec2_backend.associate_iam_instance_profile(
            instance_id, iam_instance_profile_name, iam_instance_profile_arn
        )
        result = {"IamInstanceProfileAssociation": iam_association}
        return ActionResult(result)

    def describe_iam_instance_profile_associations(self) -> ActionResult:
        association_ids = self._get_param("AssociationIds", [])
        filters = self._filters_from_querystring()
        max_items = self._get_param("MaxItems")
        next_token = self._get_param("NextToken")
        (
            iam_associations,
            next_token,
        ) = self.ec2_backend.describe_iam_instance_profile_associations(
            association_ids, filters, max_items, next_token
        )
        result = {
            "IamInstanceProfileAssociations": iam_associations,
            "NextToken": next_token,
        }
        return ActionResult(result)

    def disassociate_iam_instance_profile(self) -> ActionResult:
        association_id = self._get_param("AssociationId")
        iam_association = self.ec2_backend.disassociate_iam_instance_profile(
            association_id
        )
        result = {"IamInstanceProfileAssociation": iam_association}
        return ActionResult(result)

    def replace_iam_instance_profile_association(self) -> ActionResult:
        association_id = self._get_param("AssociationId")
        iam_instance_profile_name = self._get_param("IamInstanceProfile.Name")
        iam_instance_profile_arn = self._get_param("IamInstanceProfile.Arn")
        iam_association = self.ec2_backend.replace_iam_instance_profile_association(
            association_id, iam_instance_profile_name, iam_instance_profile_arn
        )
        result = {"IamInstanceProfileAssociation": iam_association}
        return ActionResult(result)



from typing import Any, Optional

# AWS Docs: https://docs.aws.amazon.com/elasticloadbalancing/latest/application/describe-ssl-policies.html

SSL_POLICIES = {
    "ELBSecurityPolicy-TLS13-1-3-2021-06": {
        "name": "ELBSecurityPolicy-TLS13-1-3-2021-06",
        "ciphers": [
            "TLS_AES_128_GCM_SHA256",
            "TLS_AES_256_GCM_SHA384",
            "TLS_CHACHA20_POLY1305_SHA256",
        ],
        "ssl_protocols": ["TLSv1.3"],
    },
    "ELBSecurityPolicy-TLS13-1-3-PQ-2025-09": {
        "name": "ELBSecurityPolicy-TLS13-1-3-PQ-2025-09",
        "ciphers": [
            "TLS_AES_128_GCM_SHA256",
            "TLS_AES_256_GCM_SHA384",
            "TLS_CHACHA20_POLY1305_SHA256",
        ],
        "ssl_protocols": ["TLSv1.3"],
    },
    "ELBSecurityPolicy-TLS13-1-2-2021-06": {
        "name": "ELBSecurityPolicy-TLS13-1-2-2021-06",
        "ciphers": [
            "TLS_AES_128_GCM_SHA256",
            "TLS_AES_256_GCM_SHA384",
            "TLS_CHACHA20_POLY1305_SHA256",
            "ECDHE-ECDSA-AES128-GCM-SHA256",
            "ECDHE-RSA-AES128-GCM-SHA256",
            "ECDHE-ECDSA-AES128-SHA256",
            "ECDHE-RSA-AES128-SHA256",
            "ECDHE-ECDSA-AES256-GCM-SHA384",
            "ECDHE-RSA-AES256-GCM-SHA384",
            "ECDHE-ECDSA-AES256-SHA384",
            "ECDHE-RSA-AES256-SHA384",
        ],
        "ssl_protocols": ["TLSv1.3", "TLSv1.2"],
    },
    "ELBSecurityPolicy-TLS13-1-2-PQ-2025-09": {
        "name": "ELBSecurityPolicy-TLS13-1-2-PQ-2025-09",
        "ciphers": [
            "TLS_AES_128_GCM_SHA256",
            "TLS_AES_256_GCM_SHA384",
            "TLS_CHACHA20_POLY1305_SHA256",
            "ECDHE-ECDSA-AES128-GCM-SHA256",
            "ECDHE-RSA-AES128-GCM-SHA256",
            "ECDHE-ECDSA-AES128-SHA256",
            "ECDHE-RSA-AES128-SHA256",
            "ECDHE-ECDSA-AES256-GCM-SHA384",
            "ECDHE-RSA-AES256-GCM-SHA384",
            "ECDHE-ECDSA-AES256-SHA384",
            "ECDHE-RSA-AES256-SHA384",
        ],
        "ssl_protocols": ["TLSv1.3", "TLSv1.2"],
    },
    "ELBSecurityPolicy-TLS13-1-2-Res-2021-06": {
        "name": "ELBSecurityPolicy-TLS13-1-2-Res-2021-06",
        "ciphers": [
            "TLS_AES_128_GCM_SHA256",
            "TLS_AES_256_GCM_SHA384",
            "TLS_CHACHA20_POLY1305_SHA256",
            "ECDHE-ECDSA-AES128-GCM-SHA256",
            "ECDHE-RSA-AES128-GCM-SHA256",
            "ECDHE-ECDSA-AES256-GCM-SHA384",
            "ECDHE-RSA-AES256-GCM-SHA384",
        ],
        "ssl_protocols": ["TLSv1.3", "TLSv1.2"],
    },
    "ELBSecurityPolicy-TLS13-1-2-Res-PQ-2025-09": {
        "name": "ELBSecurityPolicy-TLS13-1-2-Res-PQ-2025-09",
        "ciphers": [
            "TLS_AES_128_GCM_SHA256",
            "TLS_AES_256_GCM_SHA384",
            "TLS_CHACHA20_POLY1305_SHA256",
            "ECDHE-ECDSA-AES128-GCM-SHA256",
            "ECDHE-RSA-AES128-GCM-SHA256",
            "ECDHE-ECDSA-AES256-GCM-SHA384",
            "ECDHE-RSA-AES256-GCM-SHA384",
        ],
        "ssl_protocols": ["TLSv1.3", "TLSv1.2"],
    },
    "ELBSecurityPolicy-TLS13-1-2-Ext2-2021-06": {
        "name": "ELBSecurityPolicy-TLS13-1-2-Ext2-2021-06",
        "ciphers": [
            "TLS_AES_128_GCM_SHA256",
            "TLS_AES_256_GCM_SHA384",
            "TLS_CHACHA20_POLY1305_SHA256",
            "ECDHE-ECDSA-AES128-GCM-SHA256",
            "ECDHE-RSA-AES128-GCM-SHA256",
            "ECDHE-ECDSA-AES128-SHA256",
            "ECDHE-RSA-AES128-SHA256",
            "ECDHE-ECDSA-AES128-SHA",
            "ECDHE-RSA-AES128-SHA",
            "ECDHE-ECDSA-AES256-GCM-SHA384",
            "ECDHE-RSA-AES256-GCM-SHA384",
            "ECDHE-ECDSA-AES256-SHA384",
            "ECDHE-RSA-AES256-SHA384",
            "ECDHE-ECDSA-AES256-SHA",
            "ECDHE-RSA-AES256-SHA",
            "AES128-GCM-SHA256",
            "AES128-SHA256",
            "AES128-SHA",
            "AES256-GCM-SHA384",
            "AES256-SHA256",
            "AES256-SHA",
        ],
        "ssl_protocols": ["TLSv1.3", "TLSv1.2"],
    },
    "ELBSecurityPolicy-TLS13-1-2-Ext2-PQ-2025-09": {
        "name": "ELBSecurityPolicy-TLS13-1-2-Ext2-PQ-2025-09",
        "ciphers": [
            "TLS_AES_128_GCM_SHA256",
            "TLS_AES_256_GCM_SHA384",
            "TLS_CHACHA20_POLY1305_SHA256",
            "ECDHE-ECDSA-AES128-GCM-SHA256",
            "ECDHE-RSA-AES128-GCM-SHA256",
            "ECDHE-ECDSA-AES128-SHA256",
            "ECDHE-RSA-AES128-SHA256",
            "ECDHE-ECDSA-AES128-SHA",
            "ECDHE-RSA-AES128-SHA",
            "ECDHE-ECDSA-AES256-GCM-SHA384",
            "ECDHE-RSA-AES256-GCM-SHA384",
            "ECDHE-ECDSA-AES256-SHA384",
            "ECDHE-RSA-AES256-SHA384",
            "ECDHE-ECDSA-AES256-SHA",
            "ECDHE-RSA-AES256-SHA",
            "AES128-GCM-SHA256",
            "AES128-SHA256",
            "AES128-SHA",
            "AES256-GCM-SHA384",
            "AES256-SHA256",
            "AES256-SHA",
        ],
        "ssl_protocols": ["TLSv1.3", "TLSv1.2"],
    },
    "ELBSecurityPolicy-TLS13-1-2-Ext1-2021-06": {
        "name": "ELBSecurityPolicy-TLS13-1-2-Ext1-2021-06",
        "ciphers": [
            "TLS_AES_128_GCM_SHA256",
            "TLS_AES_256_GCM_SHA384",
            "TLS_CHACHA20_POLY1305_SHA256",
            "ECDHE-ECDSA-AES128-GCM-SHA256",
            "ECDHE-RSA-AES128-GCM-SHA256",
            "ECDHE-ECDSA-AES128-SHA256",
            "ECDHE-RSA-AES128-SHA256",
            "ECDHE-ECDSA-AES256-GCM-SHA384",
            "ECDHE-RSA-AES256-GCM-SHA384",
            "ECDHE-ECDSA-AES256-SHA384",
            "ECDHE-RSA-AES256-SHA384",
            "AES128-GCM-SHA256",
            "AES128-SHA256",
            "AES256-GCM-SHA384",
            "AES256-SHA256",
        ],
        "ssl_protocols": ["TLSv1.3", "TLSv1.2"],
    },
    "ELBSecurityPolicy-TLS13-1-2-Ext1-PQ-2025-09": {
        "name": "ELBSecurityPolicy-TLS13-1-2-Ext1-PQ-2025-09",
        "ciphers": [
            "TLS_AES_128_GCM_SHA256",
            "TLS_AES_256_GCM_SHA384",
            "TLS_CHACHA20_POLY1305_SHA256",
            "ECDHE-ECDSA-AES128-GCM-SHA256",
            "ECDHE-RSA-AES128-GCM-SHA256",
            "ECDHE-ECDSA-AES128-SHA256",
            "ECDHE-RSA-AES128-SHA256",
            "ECDHE-ECDSA-AES128-SHA",
            "ECDHE-RSA-AES128-SHA",
            "ECDHE-ECDSA-AES256-GCM-SHA384",
            "ECDHE-RSA-AES256-GCM-SHA384",
            "ECDHE-ECDSA-AES256-SHA384",
            "ECDHE-RSA-AES256-SHA384",
            "ECDHE-ECDSA-AES256-SHA",
            "ECDHE-RSA-AES256-SHA",
            "AES128-GCM-SHA256",
            "AES128-SHA256",
            "AES128-SHA",
            "AES256-GCM-SHA384",
            "AES256-SHA256",
            "AES256-SHA",
        ],
        "ssl_protocols": ["TLSv1.3", "TLSv1.2"],
    },
    "ELBSecurityPolicy-TLS13-1-1-2021-06": {
        "name": "ELBSecurityPolicy-TLS13-1-1-2021-06",
        "ciphers": [
            "TLS_AES_128_GCM_SHA256",
            "TLS_AES_256_GCM_SHA384",
            "TLS_CHACHA20_POLY1305_SHA256",
            "ECDHE-ECDSA-AES128-GCM-SHA256",
            "ECDHE-RSA-AES128-GCM-SHA256",
            "ECDHE-ECDSA-AES128-SHA256",
            "ECDHE-RSA-AES128-SHA256",
            "ECDHE-ECDSA-AES128-SHA",
            "ECDHE-RSA-AES128-SHA",
            "ECDHE-ECDSA-AES256-GCM-SHA384",
            "ECDHE-RSA-AES256-GCM-SHA384",
            "ECDHE-ECDSA-AES256-SHA384",
            "ECDHE-RSA-AES256-SHA384",
            "ECDHE-ECDSA-AES256-SHA",
            "ECDHE-RSA-AES256-SHA",
            "AES128-GCM-SHA256",
            "AES128-SHA256",
            "AES128-SHA",
            "AES256-GCM-SHA384",
            "AES256-SHA256",
            "AES256-SHA",
        ],
        "ssl_protocols": ["TLSv1.3", "TLSv1.2", "TLSv1.1"],
    },
    "ELBSecurityPolicy-TLS13-1-0-2021-06": {
        "name": "ELBSecurityPolicy-TLS13-1-0-2021-06",
        "ciphers": [
            "TLS_AES_128_GCM_SHA256",
            "TLS_AES_256_GCM_SHA384",
            "TLS_CHACHA20_POLY1305_SHA256",
            "ECDHE-ECDSA-AES128-GCM-SHA256",
            "ECDHE-RSA-AES128-GCM-SHA256",
            "ECDHE-ECDSA-AES128-SHA256",
            "ECDHE-RSA-AES128-SHA256",
            "ECDHE-ECDSA-AES128-SHA",
            "ECDHE-RSA-AES128-SHA",
            "ECDHE-ECDSA-AES256-GCM-SHA384",
            "ECDHE-RSA-AES256-GCM-SHA384",
            "ECDHE-ECDSA-AES256-SHA384",
            "ECDHE-RSA-AES256-SHA384",
            "ECDHE-ECDSA-AES256-SHA",
            "ECDHE-RSA-AES256-SHA",
            "AES128-GCM-SHA256",
            "AES128-SHA256",
            "AES128-SHA",
            "AES256-GCM-SHA384",
            "AES256-SHA256",
            "AES256-SHA",
        ],
        "ssl_protocols": ["TLSv1.3", "TLSv1.2", "TLSv1.1", "TLSv1"],
    },
    "ELBSecurityPolicy-TLS13-1-0-PQ-2025-09": {
        "name": "ELBSecurityPolicy-TLS13-1-0-PQ-2025-09",
        "ciphers": [
            "TLS_AES_128_GCM_SHA256",
            "TLS_AES_256_GCM_SHA384",
            "TLS_CHACHA20_POLY1305_SHA256",
            "ECDHE-ECDSA-AES128-GCM-SHA256",
            "ECDHE-RSA-AES128-GCM-SHA256",
            "ECDHE-ECDSA-AES128-SHA256",
            "ECDHE-RSA-AES128-SHA256",
            "ECDHE-ECDSA-AES128-SHA",
            "ECDHE-RSA-AES128-SHA",
            "ECDHE-ECDSA-AES256-GCM-SHA384",
            "ECDHE-RSA-AES256-GCM-SHA384",
            "ECDHE-ECDSA-AES256-SHA384",
            "ECDHE-RSA-AES256-SHA384",
            "ECDHE-ECDSA-AES256-SHA",
            "ECDHE-RSA-AES256-SHA",
            "AES128-GCM-SHA256",
            "AES128-SHA256",
            "AES128-SHA",
            "AES256-GCM-SHA384",
            "AES256-SHA256",
            "AES256-SHA",
        ],
        "ssl_protocols": ["TLSv1.3", "TLSv1.2", "TLSv1.1", "TLSv1"],
    },
    "ELBSecurityPolicy-TLS-1-2-Ext-2018-06": {
        "name": "ELBSecurityPolicy-TLS-1-2-Ext-2018-06",
        "ciphers": [
            "ECDHE-ECDSA-AES128-GCM-SHA256",
            "ECDHE-RSA-AES128-GCM-SHA256",
            "ECDHE-ECDSA-AES128-SHA256",
            "ECDHE-RSA-AES128-SHA256",
            "ECDHE-ECDSA-AES128-SHA",
            "ECDHE-RSA-AES128-SHA",
            "ECDHE-ECDSA-AES256-GCM-SHA384",
            "ECDHE-RSA-AES256-GCM-SHA384",
            "ECDHE-ECDSA-AES256-SHA384",
            "ECDHE-RSA-AES256-SHA384",
            "ECDHE-ECDSA-AES256-SHA",
            "ECDHE-RSA-AES256-SHA",
            "AES128-GCM-SHA256",
            "AES128-SHA256",
            "AES128-SHA",
            "AES256-GCM-SHA384",
            "AES256-SHA256",
            "AES256-SHA",
        ],
        "ssl_protocols": ["TLSv1.2"],
    },
    "ELBSecurityPolicy-TLS-1-2-2017-01": {
        "name": "ELBSecurityPolicy-TLS-1-2-2017-01",
        "ciphers": [
            "ECDHE-ECDSA-AES128-GCM-SHA256",
            "ECDHE-RSA-AES128-GCM-SHA256",
            "ECDHE-ECDSA-AES128-SHA256",
            "ECDHE-RSA-AES128-SHA256",
            "ECDHE-ECDSA-AES256-GCM-SHA384",
            "ECDHE-RSA-AES256-GCM-SHA384",
            "ECDHE-ECDSA-AES256-SHA384",
            "ECDHE-RSA-AES256-SHA384",
            "AES128-GCM-SHA256",
            "AES128-SHA256",
            "AES256-GCM-SHA384",
            "AES256-SHA256",
        ],
        "ssl_protocols": ["TLSv1.2"],
    },
    "ELBSecurityPolicy-TLS-1-1-2017-01": {
        "name": "ELBSecurityPolicy-TLS-1-1-2017-01",
        "ciphers": [
            "ECDHE-ECDSA-AES128-GCM-SHA256",
            "ECDHE-RSA-AES128-GCM-SHA256",
            "ECDHE-ECDSA-AES128-SHA256",
            "ECDHE-RSA-AES128-SHA256",
            "ECDHE-ECDSA-AES128-SHA",
            "ECDHE-RSA-AES128-SHA",
            "ECDHE-ECDSA-AES256-GCM-SHA384",
            "ECDHE-RSA-AES256-GCM-SHA384",
            "ECDHE-ECDSA-AES256-SHA384",
            "ECDHE-RSA-AES256-SHA384",
            "ECDHE-ECDSA-AES256-SHA",
            "ECDHE-RSA-AES256-SHA",
            "AES128-GCM-SHA256",
            "AES128-SHA256",
            "AES128-SHA",
            "AES256-GCM-SHA384",
            "AES256-SHA256",
            "AES256-SHA",
        ],
        "ssl_protocols": ["TLSv1.2", "TLSv1.1"],
    },
    "ELBSecurityPolicy-2016-08": {
        "name": "ELBSecurityPolicy-2016-08",
        "ciphers": [
            "ECDHE-ECDSA-AES128-GCM-SHA256",
            "ECDHE-RSA-AES128-GCM-SHA256",
            "ECDHE-ECDSA-AES128-SHA256",
            "ECDHE-RSA-AES128-SHA256",
            "ECDHE-ECDSA-AES128-SHA",
            "ECDHE-RSA-AES128-SHA",
            "ECDHE-ECDSA-AES256-GCM-SHA384",
            "ECDHE-RSA-AES256-GCM-SHA384",
            "ECDHE-ECDSA-AES256-SHA384",
            "ECDHE-RSA-AES256-SHA384",
            "ECDHE-ECDSA-AES256-SHA",
            "ECDHE-RSA-AES256-SHA",
            "AES128-GCM-SHA256",
            "AES128-SHA256",
            "AES128-SHA",
            "AES256-GCM-SHA384",
            "AES256-SHA256",
            "AES256-SHA",
        ],
        "ssl_protocols": ["TLSv1.2", "TLSv1.1", "TLSv1"],
    },
}


def build_policy(name: str, data: dict[str, Any]) -> dict[str, Optional[object]]:
    # Adding priority numbers to ciphers and restructuring output to required format
    ciphers = [
        {"Name": c, "Priority": i + 1} for i, c in enumerate(data.get("ciphers", []))
    ]
    return {
        "Name": name,
        "SslProtocols": data.get("ssl_protocols", []),
        "Ciphers": ciphers,
    }



import abc
from typing import Any, Final

from moto.stepfunctions.parser.asl.component.common.assign.assign_template_value import (
    AssignTemplateValue,
)
from moto.stepfunctions.parser.asl.component.common.string.string_expression import (
    StringJSONata,
)
from moto.stepfunctions.parser.asl.eval.environment import Environment


class AssignTemplateValueTerminal(AssignTemplateValue, abc.ABC): ...


class AssignTemplateValueTerminalLit(AssignTemplateValueTerminal):
    value: Final[Any]

    def __init__(self, value: Any):
        super().__init__()
        self.value = value

    def _eval_body(self, env: Environment) -> None:
        env.stack.append(self.value)


class AssignTemplateValueTerminalStringJSONata(AssignTemplateValueTerminal):
    string_jsonata: Final[StringJSONata]

    def __init__(self, string_jsonata: StringJSONata):
        super().__init__()
        self.string_jsonata = string_jsonata

    def _eval_body(self, env: Environment) -> None:
        self.string_jsonata.eval(env=env)



"""S3VectorsBackend class with methods for supported APIs."""

from collections.abc import Iterable
from typing import Any, Literal, Optional, TypedDict

from moto.core.base_backend import BackendDict, BaseBackend
from moto.core.common_models import BaseModel
from moto.utilities.arns import parse_arn
from moto.utilities.utils import PARTITION_NAMES

from .exceptions import (
    IndexNotFound,
    VectorBucketAlreadyExists,
    VectorBucketNotEmpty,
    VectorBucketNotFound,
    VectorBucketPolicyNotFound,
    VectorWrongDimension,
)
from .utils import create_vector_bucket_arn


class VectorData(TypedDict):
    float32: list[float]


class VectorType(TypedDict, total=False):
    key: str
    data: VectorData
    metadata: Any


class Vector(BaseModel):
    def __init__(self, key: str, data: VectorData, metadata: Any):
        self.key = key
        self.data = data
        self.metadata = metadata

    def to_dict(self, return_data: bool, return_metadata: bool) -> VectorType:
        return (
            {"key": self.key}  # type: ignore[return-value]
            | ({"data": self.data} if return_data else {})
            | ({"metadata": self.metadata} if return_metadata else {})
        )

    @staticmethod
    def from_dict(dct: VectorType) -> "Vector":
        return Vector(
            key=dct["key"], data=dct["data"], metadata=dct.get("metadata", {})
        )


class Index(BaseModel):
    def __init__(
        self,
        bucket: "VectorBucket",
        name: str,
        dimension: int,
        data_type: Literal["float32"],
        distance_metric: str,
    ):
        self.vectorBucketName = bucket.vector_bucket_name
        self.index_name = name
        self.index_arn = f"{bucket.vector_bucket_arn}/index/{name}"
        self.dimension = dimension
        self.data_type = data_type
        self.distance_metric = distance_metric

        self._bucket = bucket
        self.vectors: dict[str, Vector] = {}


class VectorBucket(BaseModel):
    def __init__(
        self,
        arn: str,
        name: str,
        encryption_configuration: dict[str, str],
    ):
        self.vector_bucket_name = name
        self.vector_bucket_arn = arn
        self.encryption_configuration = encryption_configuration or {
            "sseType": "AES256"
        }

        self.indexes: dict[str, Index] = {}
        self.policy: Optional[str] = None


class S3VectorsBackend(BaseBackend):
    """Implementation of S3Vectors APIs."""

    def __init__(self, region_name: str, account_id: str):
        super().__init__(region_name, account_id)
        self.vector_buckets: dict[str, VectorBucket] = {}

    def create_vector_bucket(
        self,
        region: str,
        vector_bucket_name: str,
        encryption_configuration: dict[str, str],
    ) -> None:
        vector_bucket_arn = create_vector_bucket_arn(
            self.account_id, region, name=vector_bucket_name
        )
        if vector_bucket_arn in self.vector_buckets:
            raise VectorBucketAlreadyExists
        vector_bucket = VectorBucket(
            arn=vector_bucket_arn,
            name=vector_bucket_name,
            encryption_configuration=encryption_configuration,
        )
        self.vector_buckets[vector_bucket.vector_bucket_arn] = vector_bucket

    def get_vector_bucket(
        self,
        vector_bucket_name: Optional[str] = None,
        vector_bucket_arn: Optional[str] = None,
    ) -> VectorBucket:
        if vector_bucket_name:
            for vector_bucket in self.vector_buckets.values():
                if vector_bucket.vector_bucket_name == vector_bucket_name:
                    return vector_bucket
        if vector_bucket_arn and (bucket := self.vector_buckets.get(vector_bucket_arn)):
            return bucket
        raise VectorBucketNotFound

    def delete_vector_bucket(self, vector_bucket_name: str) -> None:
        if vector_bucket_name:
            bucket = self.get_vector_bucket(vector_bucket_name=vector_bucket_name)
            if bucket.indexes:
                raise VectorBucketNotEmpty
            self.vector_buckets.pop(bucket.vector_bucket_arn, None)

    def list_vector_buckets(self, prefix: Optional[str]) -> list[VectorBucket]:
        return [
            bucket
            for bucket in self.vector_buckets.values()
            if not prefix or bucket.vector_bucket_name.startswith(prefix)
        ]

    def create_index(
        self,
        vector_bucket_name: str,
        vector_bucket_arn: str,
        index_name: str,
        data_type: Literal["float32"],
        dimension: int,
        distance_metric: str,
    ) -> None:
        bucket = self.get_vector_bucket(
            vector_bucket_name=vector_bucket_name, vector_bucket_arn=vector_bucket_arn
        )
        index = Index(
            bucket=bucket,
            name=index_name,
            data_type=data_type,
            dimension=dimension,
            distance_metric=distance_metric,
        )
        bucket.indexes[index.index_arn] = index

    def delete_index(
        self, vector_bucket_name: str, index_name: str, index_arn: str
    ) -> None:
        index = self.get_index(vector_bucket_name, index_name, index_arn)
        index._bucket.indexes.pop(index.index_arn)

    def get_index(
        self, vector_bucket_name: str, index_name: str, index_arn: str
    ) -> Index:
        if index_arn:
            vector_bucket_name, _, index_name = parse_arn(index_arn).resource_id.split(
                "/"
            )
        try:
            bucket = self.get_vector_bucket(
                vector_bucket_name=vector_bucket_name, vector_bucket_arn=None
            )
            for index in bucket.indexes.values():
                if index.index_name == index_name:
                    return index
        except VectorBucketNotFound:
            pass
        raise IndexNotFound

    def list_indexes(
        self, vector_bucket_name: str, vector_bucket_arn: str
    ) -> list[Index]:
        """Pagination is not yet implemented. The prefix-parameter is also not yet implemented."""
        bucket = self.get_vector_bucket(
            vector_bucket_name, vector_bucket_arn=vector_bucket_arn
        )
        return list(bucket.indexes.values())

    def delete_vector_bucket_policy(
        self, vector_bucket_name: str, vector_bucket_arn: str
    ) -> None:
        bucket = self.get_vector_bucket(
            vector_bucket_name, vector_bucket_arn=vector_bucket_arn
        )
        bucket.policy = None

    def get_vector_bucket_policy(
        self, vector_bucket_name: str, vector_bucket_arn: str
    ) -> str:
        bucket = self.get_vector_bucket(
            vector_bucket_name, vector_bucket_arn=vector_bucket_arn
        )
        if not bucket.policy:
            raise VectorBucketPolicyNotFound
        return bucket.policy

    def put_vector_bucket_policy(
        self, vector_bucket_name: str, vector_bucket_arn: str, policy: str
    ) -> None:
        bucket = self.get_vector_bucket(
            vector_bucket_name, vector_bucket_arn=vector_bucket_arn
        )
        bucket.policy = policy

    def put_vectors(
        self,
        vector_bucket_name: str,
        index_name: str,
        index_arn: str,
        vectors: list[VectorType],
    ) -> None:
        index = self.get_index(
            vector_bucket_name, index_name=index_name, index_arn=index_arn
        )

        for vector in vectors:
            provided = len(vector["data"][index.data_type])  # type: ignore[literal-required]
            if provided != index.dimension:
                raise VectorWrongDimension(
                    key=vector["key"], actual=index.dimension, provided=provided
                )

        for vector in vectors:
            index.vectors[vector["key"]] = Vector.from_dict(vector)

    def get_vectors(
        self, vector_bucket_name: str, index_name: str, index_arn: str, keys: list[str]
    ) -> list[Vector]:
        index = self.get_index(
            vector_bucket_name, index_name=index_name, index_arn=index_arn
        )
        return [index.vectors[name] for name in index.vectors if name in keys]

    def list_vectors(
        self, vector_bucket_name: str, index_name: str, index_arn: str
    ) -> Iterable[Vector]:
        """
        Pagination is not yet implemented
        Segmentation is not yet implemented
        """
        index = self.get_index(
            vector_bucket_name, index_name=index_name, index_arn=index_arn
        )
        return index.vectors.values()

    def delete_vectors(
        self, vector_bucket_name: str, index_name: str, index_arn: str, keys: list[str]
    ) -> None:
        index = self.get_index(
            vector_bucket_name, index_name=index_name, index_arn=index_arn
        )
        for key in keys:
            index.vectors.pop(key, None)


s3vectors_backends = BackendDict(
    S3VectorsBackend,
    "s3vectors",
    use_boto3_regions=False,
    additional_regions=PARTITION_NAMES,
)



from moto.core.exceptions import ServiceException


class ECSException(ServiceException):
    pass


class ServiceNotFoundException(ECSException):
    def __init__(self) -> None:
        super().__init__("ServiceNotFoundException", "Service not found.")


class TaskDefinitionNotFoundException(ECSException):
    def __init__(self) -> None:
        super().__init__(
            "ClientException",
            "Unable to describe task definition.",
        )


class RevisionNotFoundException(ECSException):
    def __init__(self) -> None:
        super().__init__("ClientException", "Revision is missing.")


class TaskSetNotFoundException(ECSException):
    def __init__(self) -> None:
        super().__init__(
            "ClientException",
            "The specified task set does not exist.",
        )


class ClusterNotFoundException(ECSException):
    def __init__(self) -> None:
        super().__init__("ClusterNotFoundException", "Cluster not found.")


class EcsClientException(ECSException):
    def __init__(self, message: str):
        super().__init__("ClientException", message)


class InvalidParameterException(ECSException):
    code = "InvalidParameterException"

    def __init__(self, message: str):
        super().__init__(message)


class UnknownAccountSettingException(InvalidParameterException):
    def __init__(self) -> None:
        super().__init__(
            "unknown should be one of [serviceLongArnFormat,taskLongArnFormat,containerInstanceLongArnFormat,containerLongArnFormat,awsvpcTrunking,containerInsights,dualStackIPv6]"
        )


class TaskDefinitionMemoryError(ECSException):
    def __init__(self, container_name: str) -> None:
        super().__init__(
            "ClientException",
            f"Invalid setting for container '{container_name}'. At least one of 'memory' or 'memoryReservation' must be specified.",
        )


class TaskDefinitionMissingPropertyError(ECSException):
    def __init__(self, missing_prop: str) -> None:
        super().__init__(
            "ClientException",
            f"Container.{missing_prop} should not be null or empty.",
        )



import functools
import inspect
import itertools
import os
import re
import unittest
from collections.abc import Callable
from contextlib import AbstractContextManager
from threading import Lock
from types import FunctionType
from typing import (
    TYPE_CHECKING,
    Any,
    Optional,
    TypeVar,
)
from unittest.mock import patch

import boto3
import botocore
import responses
from botocore.config import Config
from botocore.handlers import BUILTIN_HANDLERS

import moto.backend_index as backend_index
from moto import settings

from .base_backend import BackendDict
from .botocore_stubber import BotocoreStubber
from .config import DefaultConfig, default_user_config, mock_credentials
from .custom_responses_mock import (
    CallbackResponse,
    get_response_mock,
    not_implemented_callback,
    reset_responses_mock,
)
from .model_instances import reset_model_data

if TYPE_CHECKING:
    from typing_extensions import ParamSpec

    P = ParamSpec("P")


DEFAULT_ACCOUNT_ID = "123456789012"
T = TypeVar("T")


class MockAWS(AbstractContextManager["MockAWS"]):
    _nested_count = 0
    _mocks_active = False
    _mock_init_lock = Lock()

    def __init__(self, config: Optional[DefaultConfig] = None) -> None:
        self._fake_creds = {
            "AWS_ACCESS_KEY_ID": "FOOBARKEY",
            "AWS_SECRET_ACCESS_KEY": "FOOBARSECRET",
        }
        self._orig_creds: dict[str, Optional[str]] = {}
        self._default_session_mock = patch("boto3.DEFAULT_SESSION", None)
        current_user_config = default_user_config.copy()
        current_user_config.update(config or {})
        self._user_config_mock = patch.dict(default_user_config, current_user_config)

    def __call__(
        self, func: "Callable[P, T]", reset: bool = True, remove_data: bool = True
    ) -> "Callable[P, T]":
        if inspect.isclass(func):
            return self._decorate_class(func)
        return self._decorate_callable(func, reset, remove_data)

    def __enter__(self) -> "MockAWS":
        self.start()
        return self

    def __exit__(self, *args: Any) -> None:
        self.stop()

    def start(self, reset: bool = True) -> None:
        with MockAWS._mock_init_lock:
            self._user_config_mock.start()
            if mock_credentials():
                self._mock_env_variables()
            if not self.__class__._mocks_active:
                if default_user_config.get("core", {}).get("reset_boto3_session", True):
                    self._default_session_mock.start()
                self.__class__._mocks_active = True

            self.__class__._nested_count += 1

            if self.__class__._nested_count == 1:
                self._enable_patching(reset=reset)

    def stop(self, remove_data: bool = True) -> None:
        with MockAWS._mock_init_lock:
            self.__class__._nested_count -= 1

            if self.__class__._nested_count < 0:
                raise RuntimeError("Called stop() before start().")

            if mock_credentials():
                self._unmock_env_variables()

            if self.__class__._nested_count == 0:
                if self.__class__._mocks_active:
                    if default_user_config.get("core", {}).get(
                        "reset_boto3_session", True
                    ):
                        self._default_session_mock.stop()
                    self._user_config_mock.stop()
                    self.__class__._mocks_active = False
                self._disable_patching(remove_data)

    def _decorate_callable(
        self, func: "Callable[P, T]", reset: bool, remove_data: bool
    ) -> "Callable[P, T]":
        def wrapper(*args: Any, **kwargs: Any) -> T:
            self.start(reset=reset)
            try:
                result = func(*args, **kwargs)
            finally:
                self.stop(remove_data=remove_data)
            return result

        functools.update_wrapper(wrapper, func)
        wrapper.__wrapped__ = func  # type: ignore[attr-defined]
        return wrapper

    def _decorate_class(self, klass: "Callable[P, T]") -> "Callable[P, T]":
        assert inspect.isclass(klass)  # Keep mypy happy
        direct_methods = get_direct_methods_of(klass)
        defined_classes = {x for x, y in klass.__dict__.items() if inspect.isclass(y)}

        # Get a list of all userdefined superclasses
        superclasses = [
            c for c in klass.__mro__ if c not in [unittest.TestCase, object]
        ]
        # Get a list of all userdefined methods
        supermethods = list(
            itertools.chain(*[get_direct_methods_of(c) for c in superclasses])
        )
        # Check whether the user has overridden the setUp-method
        has_setup_method = (
            ("setUp" in supermethods and unittest.TestCase in klass.__mro__)
            or "setup" in supermethods
            or "setup_method" in supermethods
        )

        for attr in itertools.chain(direct_methods, defined_classes):
            if attr.startswith("_"):
                continue

            attr_value = getattr(klass, attr)
            if not callable(attr_value):
                continue
            if not hasattr(attr_value, "__name__"):
                continue

            # Check if this is a classmethod. If so, skip patching
            if inspect.ismethod(attr_value) and attr_value.__self__ is klass:
                continue

            # Check if this is a staticmethod. If so, skip patching
            for cls in inspect.getmro(klass):
                if attr_value.__name__ not in cls.__dict__:
                    continue
                bound_attr_value = cls.__dict__[attr_value.__name__]
                if not isinstance(bound_attr_value, staticmethod):
                    break
            else:
                # It is a staticmethod, skip patching
                continue

            try:
                # Special case for UnitTests-class
                is_test_method = attr.startswith(unittest.TestLoader.testMethodPrefix)
                should_reset = False
                should_remove_data = False
                if attr in ["setUp", "setup_method"]:
                    should_reset = True
                elif not has_setup_method and is_test_method:
                    should_reset = True
                    should_remove_data = True
                else:
                    # Method is unrelated to the test setup
                    # Method is a test, but was already reset while executing the setUp-method
                    pass
                kwargs = {"reset": should_reset, "remove_data": should_remove_data}
                setattr(klass, attr, self(attr_value, **kwargs))
            except TypeError:
                # Sometimes we can't set this for built-in types
                continue
        return klass

    def _mock_env_variables(self) -> None:
        # "Mock" the AWS credentials as they can't be mocked in Botocore currently
        for k, v in self._fake_creds.items():
            self._orig_creds[k] = os.environ.get(k, None)
            os.environ[k] = v

    def _unmock_env_variables(self) -> None:
        for k, v in self._orig_creds.items():
            if v:
                os.environ[k] = v
            else:
                del os.environ[k]

    def reset(self) -> None:
        BackendDict.reset()
        reset_responses_mock(responses_mock)

    def _enable_patching(self, reset: bool = True) -> None:
        botocore_stubber.enabled = True
        if reset:
            self.reset()
        responses_mock.start()

        for method in RESPONSES_METHODS:
            for _, pattern in backend_index.backend_url_patterns:
                responses_mock.add(
                    CallbackResponse(
                        method=method,
                        url=pattern,
                        callback=botocore_stubber.process_request,
                    )
                )
            responses_mock.add(
                CallbackResponse(
                    method=method,
                    url=re.compile(r"https?://.+\.amazonaws.com/.*"),
                    callback=not_implemented_callback,
                )
            )

    def _disable_patching(self, remove_data: bool) -> None:
        botocore_stubber.enabled = False
        if remove_data:
            self.reset()
            reset_model_data()

        responses_mock.stop()


def get_direct_methods_of(klass: object) -> set[str]:
    return {
        x
        for x, y in klass.__dict__.items()
        if isinstance(y, (FunctionType, classmethod, staticmethod))
    }


RESPONSES_METHODS = [
    responses.GET,
    responses.DELETE,
    responses.HEAD,
    responses.OPTIONS,
    responses.PATCH,
    responses.POST,
    responses.PUT,
]

responses_mock = get_response_mock()

BOTOCORE_HTTP_METHODS = ["GET", "DELETE", "HEAD", "OPTIONS", "PATCH", "POST", "PUT"]


botocore_stubber = BotocoreStubber()
BUILTIN_HANDLERS.append(("before-send", botocore_stubber))


def patch_client(client: botocore.client.BaseClient) -> None:
    """
    Explicitly patch a boto3-client
    """
    """
    Adding the botocore_stubber to the BUILTIN_HANDLERS, as above, will mock everything as long as the import ordering is correct
     - user:   start mock_aws decorator
     - system: imports core.model
     - system: adds the stubber to the BUILTIN_HANDLERS
     - user:   create a boto3 client - which will use the BUILTIN_HANDLERS

    But, if for whatever reason the imports are wrong and the client is created first, it doesn't know about our stub yet
    This method can be used to tell a client that it needs to be mocked, and append the botocore_stubber after creation
    :param client:
    :return:
    """
    if isinstance(client, botocore.client.BaseClient):
        # Check if our event handler was already registered
        try:
            event_emitter = client._ruleset_resolver._event_emitter._emitter  # type: ignore[attr-defined]
            all_handlers = event_emitter._handlers._root["children"]
            handler_trie = list(all_handlers["before-send"].values())[1]
            handlers_list = handler_trie.first + handler_trie.middle + handler_trie.last
            if botocore_stubber in handlers_list:
                # No need to patch - this client already has the botocore_stubber registered
                return
        except:  # noqa: E722 Do not use bare except
            # Because we're accessing all kinds of private methods, the API may change and newer versions of botocore may throw an exception
            # One of our tests will fail if this happens (test_patch_can_be_called_on_a_mocked_client)
            # If this happens for a user, just continue and hope for the best
            #  - in 99% of the cases there are no duplicate event handlers, so it doesn't matter if the check fails
            pass

        client.meta.events.register("before-send", botocore_stubber)  # type: ignore[arg-type]
    else:
        raise Exception(f"Argument {client} should be of type boto3.client")


def patch_resource(resource: Any) -> None:
    """
    Explicitly patch a boto3-resource
    """
    if hasattr(resource, "meta") and isinstance(
        resource.meta, boto3.resources.factory.ResourceMeta
    ):
        patch_client(resource.meta.client)
    else:
        raise Exception(f"Argument {resource} should be of type boto3.resource")


def override_responses_real_send(user_mock: Optional[responses.RequestsMock]) -> None:
    """
    Moto creates it's own Responses-object responsible for intercepting AWS requests
    If a custom Responses-object is created by the user, Moto will hijack any of the pass-thru's set

    Call this method to ensure any requests unknown to Moto are passed through the custom Responses-object.

    Set the user_mock argument to None to reset this behaviour.

    Note that this is only supported from Responses>=0.24.0
    """
    if user_mock is None:
        responses_mock._real_send = responses._real_send
    else:
        responses_mock._real_send = user_mock.unbound_on_send()


class ServerModeMockAWS(MockAWS):
    _RESET_IN_PROGRESS = False

    def __init__(self, *args: Any, **kwargs: Any):
        self._test_server_mode_endpoint = settings.test_server_mode_endpoint()
        super().__init__(*args, **kwargs)

    def reset(self) -> None:
        call_reset_api = os.environ.get("MOTO_CALL_RESET_API")
        if not call_reset_api or call_reset_api.lower() != "false":
            if not ServerModeMockAWS._RESET_IN_PROGRESS:
                ServerModeMockAWS._RESET_IN_PROGRESS = True
                import requests

                requests.post(f"{self._test_server_mode_endpoint}/moto-api/reset")
                ServerModeMockAWS._RESET_IN_PROGRESS = False

    def _enable_patching(self, reset: bool = True) -> None:
        if self.__class__._nested_count == 1 and reset:
            # Just started
            self.reset()

        from boto3 import client as real_boto3_client
        from boto3 import resource as real_boto3_resource

        def fake_boto3_client(*args: Any, **kwargs: Any) -> botocore.client.BaseClient:
            region = self._get_region(*args, **kwargs)
            if region:
                if "config" in kwargs:
                    user_agent = kwargs["config"].__dict__.get("user_agent_extra") or ""
                    kwargs["config"].__dict__["user_agent_extra"] = (
                        f"{user_agent} region/{region}"
                    )
                else:
                    config = Config(user_agent_extra="region/" + region)
                    kwargs["config"] = config
            if "endpoint_url" not in kwargs:
                kwargs["endpoint_url"] = self._test_server_mode_endpoint
            return real_boto3_client(*args, **kwargs)

        def fake_boto3_resource(*args: Any, **kwargs: Any) -> Any:
            if "endpoint_url" not in kwargs:
                kwargs["endpoint_url"] = self._test_server_mode_endpoint
            return real_boto3_resource(*args, **kwargs)

        self._client_patcher = patch("boto3.client", fake_boto3_client)
        self._resource_patcher = patch("boto3.resource", fake_boto3_resource)
        self._client_patcher.start()
        self._resource_patcher.start()

    def _get_region(self, *args: Any, **kwargs: Any) -> Optional[str]:
        if "region_name" in kwargs:
            return kwargs["region_name"]
        if type(args) is tuple and len(args) == 2:
            _, region = args
            return region
        return None

    def _disable_patching(self, remove_data: bool) -> None:
        if self._client_patcher:
            self._client_patcher.stop()
            self._resource_patcher.stop()
        if remove_data:
            self.reset()


class ProxyModeMockAWS(MockAWS):
    _RESET_IN_PROGRESS = False

    def __init__(self, *args: Any, **kwargs: Any):
        self._test_proxy_mode_endpoint = settings.test_proxy_mode_endpoint()
        super().__init__(*args, **kwargs)

    def reset(self) -> None:
        call_reset_api = os.environ.get("MOTO_CALL_RESET_API")
        if not call_reset_api or call_reset_api.lower() != "false":
            if not ProxyModeMockAWS._RESET_IN_PROGRESS:
                ProxyModeMockAWS._RESET_IN_PROGRESS = True
                import requests

                requests.post(f"{self._test_proxy_mode_endpoint}/moto-api/reset")
                ProxyModeMockAWS._RESET_IN_PROGRESS = False

    def _enable_patching(self, reset: bool = True) -> None:
        if self.__class__._nested_count == 1 and reset:
            # Just started
            self.reset()

        from boto3 import client as real_boto3_client
        from boto3 import resource as real_boto3_resource

        def fake_boto3_client(*args: Any, **kwargs: Any) -> botocore.client.BaseClient:
            kwargs["verify"] = False
            proxy_endpoint = (
                f"http://localhost:{os.environ.get('MOTO_PROXY_PORT', 5005)}"
            )
            proxies = {"http": proxy_endpoint, "https": proxy_endpoint}
            if "config" in kwargs:
                kwargs["config"].__dict__["proxies"] = proxies
            else:
                config = Config(proxies=proxies)
                kwargs["config"] = config

            return real_boto3_client(*args, **kwargs)

        def fake_boto3_resource(*args: Any, **kwargs: Any) -> Any:
            kwargs["verify"] = False
            proxy_endpoint = (
                f"http://localhost:{os.environ.get('MOTO_PROXY_PORT', 5005)}"
            )
            proxies = {"http": proxy_endpoint, "https": proxy_endpoint}
            if "config" in kwargs:
                kwargs["config"].__dict__["proxies"] = proxies
            else:
                config = Config(proxies=proxies)
                kwargs["config"] = config
            return real_boto3_resource(*args, **kwargs)

        self._client_patcher = patch("boto3.client", fake_boto3_client)
        self._resource_patcher = patch("boto3.resource", fake_boto3_resource)
        self._client_patcher.start()
        self._resource_patcher.start()

    def _disable_patching(self, remove_data: bool) -> None:
        if self._client_patcher:
            self._client_patcher.stop()
            self._resource_patcher.stop()



import base64
import re
from collections import defaultdict
from typing import Any

from moto.core.responses import TYPE_RESPONSE, ActionResult, BaseResponse, EmptyResult
from moto.core.utils import camelcase_to_underscores, ensure_boolean

from .exceptions import InvalidParameterValue, SNSInvalidParameter, SNSNotFoundError
from .models import SNSBackend, sns_backends
from .utils import is_e164


def transform_tags(tags: dict[str, str]) -> list[dict[str, str]]:
    return [{"Key": key, "Value": value} for key, value in tags.items()]


def transform_attributes(attributes: dict[str, str]) -> dict[str, str]:
    # Botocore models this as dict[str, str], but responses from AWS serialize boolean values
    # as `true` not `True`, so we have to do a bit of manipulation here.
    return {
        key: str(value).lower() if isinstance(value, bool) else value
        for key, value in attributes.items()
    }


class SNSResponse(BaseResponse):
    SMS_ATTR_REGEX = re.compile(
        r"^attributes\.entry\.(?P<index>\d+)\.(?P<type>key|value)$"
    )
    OPT_OUT_PHONE_NUMBER_REGEX = re.compile(r"^\+?\d+$")
    RESPONSE_KEY_PATH_TO_TRANSFORMER = {
        "GetEndpointAttributesResponse.Attributes": transform_attributes,
        "GetPlatformApplicationAttributesResponse.Attributes": transform_attributes,
        "GetSMSAttributesResponse.Attributes": transform_attributes,
        "GetSubscriptionAttributesResponse.Attributes": transform_attributes,
        "GetTopicAttributesResponse.Attributes": transform_attributes,
        "ListEndpointsByPlatformApplicationResponse.Endpoints.Endpoint.Attributes": transform_attributes,
        "ListPlatformApplicationsResponse.PlatformApplications.PlatformApplication.Attributes": transform_attributes,
        "ListTagsForResourceResponse.Tags": transform_tags,
    }

    def __init__(self) -> None:
        super().__init__(service_name="sns")
        self.automated_parameter_parsing = True

    @property
    def backend(self) -> SNSBackend:
        return sns_backends[self.current_account][self.region]

    def _get_attributes(self) -> dict[str, Any]:
        attributes = self._get_param("Attributes", {})
        boolean_attributes = [
            "ContentBasedDeduplication",
            "Enabled",
            "FifoTopic",
            "RawMessageDelivery",
        ]
        for attribute in boolean_attributes:
            if attribute in attributes:
                attributes[attribute] = ensure_boolean(attributes[attribute])
        return attributes

    def _get_tags(self) -> dict[str, str]:
        tags = self._get_param("Tags", [])
        return {tag["Key"]: tag["Value"] for tag in tags}

    def _parse_message_attributes(self) -> dict[str, Any]:
        message_attributes = self._get_param("MessageAttributes", {})
        return self._transform_message_attributes(message_attributes)

    def _transform_message_attributes(
        self, message_attributes: dict[str, Any]
    ) -> dict[str, Any]:
        # SNS converts some key names before forwarding messages
        # DataType -> Type, StringValue -> Value, BinaryValue -> Value
        transformed_message_attributes = {}
        for name, value in message_attributes.items():
            # validation
            data_type = value["DataType"]
            if not data_type:
                raise InvalidParameterValue(
                    f"The message attribute '{name}' must contain non-empty message attribute value."
                )

            data_type_parts = data_type.split(".")
            if data_type_parts[0] not in [
                "String",
                "Binary",
                "Number",
            ]:
                raise InvalidParameterValue(
                    f"The message attribute '{name}' has an invalid message "
                    "attribute type, the set of supported type prefixes is "
                    "Binary, Number, and String."
                )

            transform_value = None
            if "StringValue" in value:
                transform_value = value["StringValue"]
                if data_type == "Number":
                    try:
                        int(transform_value)
                    except ValueError:
                        try:
                            float(transform_value)
                        except ValueError:
                            raise InvalidParameterValue(
                                "An error occurred (ParameterValueInvalid) "
                                "when calling the Publish operation: "
                                f"Could not cast message attribute '{name}' value to number."
                            )
            elif "BinaryValue" in value:
                transform_value = base64.b64encode(value["BinaryValue"]).decode()
            if transform_value == "":
                raise InvalidParameterValue(
                    f"The message attribute '{name}' must contain non-empty "
                    "message attribute value for message attribute "
                    f"type '{data_type[0]}'."
                )

            # transformation
            transformed_message_attributes[name] = {
                "Type": data_type,
                "Value": transform_value,
            }

        return transformed_message_attributes

    def create_topic(self) -> ActionResult:
        name = self._get_param("Name")
        attributes = self._get_attributes()
        tags = self._get_tags()
        topic = self.backend.create_topic(name, attributes, tags)
        result = {"TopicArn": topic.arn}
        return ActionResult(result)

    def list_topics(self) -> ActionResult:
        next_token = self._get_param("NextToken")
        topics, next_token = self.backend.list_topics(next_token=next_token)
        result = {"Topics": topics, "NextToken": next_token}
        return ActionResult(result)

    def delete_topic(self) -> ActionResult:
        topic_arn = self._get_param("TopicArn")
        self.backend.delete_topic(topic_arn)
        return EmptyResult()

    def get_topic_attributes(self) -> ActionResult:
        topic_arn = self._get_param("TopicArn")
        topic = self.backend.get_topic(topic_arn)
        attributes = {
            "Owner": topic.account_id,
            "Policy": topic.policy,
            "TopicArn": topic.arn,
            "DisplayName": topic.display_name,
            "SubscriptionsPending": topic.subscriptions_pending,
            "SubscriptionsConfirmed": topic.subscriptions_confimed,
            "SubscriptionsDeleted": topic.subscriptions_deleted,
            "DeliveryPolicy": topic.delivery_policy,
            "EffectiveDeliveryPolicy": topic.effective_delivery_policy,
        }
        if topic.kms_master_key_id:
            attributes["KmsMasterKeyId"] = topic.kms_master_key_id
        if topic.fifo_topic:
            attributes["FifoTopic"] = topic.fifo_topic
            attributes["ContentBasedDeduplication"] = topic.content_based_deduplication
        result = {"Attributes": attributes}
        return ActionResult(result)

    def set_topic_attributes(self) -> ActionResult:
        topic_arn = self._get_param("TopicArn")
        attribute_name = self._get_param("AttributeName")
        attribute_value = self._get_param("AttributeValue")
        if attribute_name == "ContentBasedDeduplication":
            attribute_value = ensure_boolean(attribute_value)
        attribute_name = camelcase_to_underscores(attribute_name)
        self.backend.set_topic_attribute(topic_arn, attribute_name, attribute_value)
        return EmptyResult()

    def subscribe(self) -> ActionResult:
        topic_arn = self._get_param("TopicArn")
        endpoint = self._get_param("Endpoint")
        protocol = self._get_param("Protocol")
        attributes = self._get_attributes()

        subscription = self.backend.subscribe(topic_arn, endpoint, protocol)

        if attributes is not None:
            # We need to set the FilterPolicyScope first, as the validation of the FilterPolicy will depend on it
            if "FilterPolicyScope" in attributes:
                filter_policy_scope = attributes.pop("FilterPolicyScope")
                self.backend.set_subscription_attributes(
                    subscription.arn, "FilterPolicyScope", filter_policy_scope
                )

            for attr_name, attr_value in attributes.items():
                self.backend.set_subscription_attributes(
                    subscription.arn, attr_name, attr_value
                )

        result = {"SubscriptionArn": subscription.arn}
        return ActionResult(result)

    def unsubscribe(self) -> ActionResult:
        subscription_arn = self._get_param("SubscriptionArn")
        self.backend.unsubscribe(subscription_arn)
        return EmptyResult()

    def list_subscriptions(self) -> ActionResult:
        next_token = self._get_param("NextToken")
        subscriptions, next_token = self.backend.list_subscriptions(
            next_token=next_token
        )
        result = {"Subscriptions": subscriptions, "NextToken": next_token}
        return ActionResult(result)

    def list_subscriptions_by_topic(self) -> ActionResult:
        topic_arn = self._get_param("TopicArn")
        next_token = self._get_param("NextToken")
        subscriptions, next_token = self.backend.list_subscriptions_by_topic(
            topic_arn, next_token=next_token
        )
        result = {"Subscriptions": subscriptions, "NextToken": next_token}
        return ActionResult(result)

    def publish(self) -> ActionResult:
        target_arn = self._get_param("TargetArn")
        topic_arn = self._get_param("TopicArn")
        phone_number = self._get_param("PhoneNumber")
        subject = self._get_param("Subject")
        message_group_id = self._get_param("MessageGroupId")
        message_deduplication_id = self._get_param("MessageDeduplicationId")
        message_structure = self._get_param("MessageStructure")

        message_attributes = self._parse_message_attributes()

        arn = None
        if phone_number is not None:
            # Check phone is correct syntax (e164)
            if not is_e164(phone_number):
                raise SNSInvalidParameter("Phone number does not meet the E164 format")
        elif target_arn is not None:
            arn = target_arn
        else:
            arn = topic_arn

        message = self._get_param("Message")

        try:
            message_id = self.backend.publish(
                message,
                arn=arn,
                phone_number=phone_number,
                subject=subject,
                message_attributes=message_attributes,
                group_id=message_group_id,
                deduplication_id=message_deduplication_id,
                message_structure=message_structure,
            )
        except ValueError as err:
            raise SNSInvalidParameter(str(err))

        result = {"MessageId": message_id}
        return ActionResult(result)

    def publish_batch(self) -> ActionResult:
        topic_arn = self._get_param("TopicArn")
        publish_batch_request_entries = self._get_param(
            "PublishBatchRequestEntries", []
        )
        for entry in publish_batch_request_entries:
            if "MessageAttributes" in entry:
                # Use the same validation/processing as the regular publish-method
                entry["MessageAttributes"] = self._transform_message_attributes(
                    entry["MessageAttributes"]
                )
        successful, failed = self.backend.publish_batch(
            topic_arn=topic_arn,
            publish_batch_request_entries=publish_batch_request_entries,
        )
        result = {"Successful": successful, "Failed": failed}
        return ActionResult(result)

    def create_platform_application(self) -> ActionResult:
        name = self._get_param("Name")
        platform = self._get_param("Platform")
        attributes = self._get_attributes()
        platform_application = self.backend.create_platform_application(
            name, platform, attributes
        )
        result = {"PlatformApplicationArn": platform_application.arn}
        return ActionResult(result)

    def get_platform_application_attributes(self) -> ActionResult:
        arn = self._get_param("PlatformApplicationArn")
        attributes = self.backend.get_platform_application_attributes(arn)
        result = {"Attributes": attributes}
        return ActionResult(result)

    def set_platform_application_attributes(self) -> ActionResult:
        arn = self._get_param("PlatformApplicationArn")
        attributes = self._get_attributes()

        self.backend.set_platform_application_attributes(arn, attributes)
        return EmptyResult()

    def list_platform_applications(self) -> ActionResult:
        applications = self.backend.list_platform_applications()
        result = {"PlatformApplications": applications}
        return ActionResult(result)

    def delete_platform_application(self) -> ActionResult:
        platform_arn = self._get_param("PlatformApplicationArn")
        self.backend.delete_platform_application(platform_arn)
        return EmptyResult()

    def create_platform_endpoint(self) -> ActionResult:
        application_arn = self._get_param("PlatformApplicationArn")
        application = self.backend.get_application(application_arn)
        custom_user_data = self._get_param("CustomUserData")
        token = self._get_param("Token")
        attributes = self._get_attributes()
        platform_endpoint = self.backend.create_platform_endpoint(
            application, custom_user_data, token, attributes
        )
        result = {"EndpointArn": platform_endpoint.arn}
        return ActionResult(result)

    def list_endpoints_by_platform_application(self) -> ActionResult:
        application_arn = self._get_param("PlatformApplicationArn")
        self.backend.get_application(application_arn)
        endpoints = self.backend.list_endpoints_by_platform_application(application_arn)
        result = {"Endpoints": endpoints}
        return ActionResult(result)

    def get_endpoint_attributes(self) -> ActionResult:
        arn = self._get_param("EndpointArn")
        attributes = self.backend.get_endpoint_attributes(arn)
        result = {"Attributes": attributes}
        return ActionResult(result)

    def set_endpoint_attributes(self) -> ActionResult:
        arn = self._get_param("EndpointArn")
        attributes = self._get_attributes()
        self.backend.set_endpoint_attributes(arn, attributes)
        return EmptyResult()

    def delete_endpoint(self) -> ActionResult:
        arn = self._get_param("EndpointArn")
        self.backend.delete_endpoint(arn)
        return EmptyResult()

    def get_subscription_attributes(self) -> ActionResult:
        arn = self._get_param("SubscriptionArn")
        attributes = self.backend.get_subscription_attributes(arn)
        result = {"Attributes": attributes}
        return ActionResult(result)

    def set_subscription_attributes(self) -> ActionResult:
        arn = self._get_param("SubscriptionArn")
        attr_name = self._get_param("AttributeName")
        attr_value = self._get_param("AttributeValue")
        if attr_name == "RawMessageDelivery":
            attr_value = ensure_boolean(attr_value)
        self.backend.set_subscription_attributes(arn, attr_name, attr_value)
        return EmptyResult()

    def set_sms_attributes(self) -> ActionResult:
        # attributes.entry.1.key
        # attributes.entry.1.value
        # to
        # 1: {key:X, value:Y}
        temp_dict: dict[str, Any] = defaultdict(dict)
        for key, value in self.querystring.items():
            match = self.SMS_ATTR_REGEX.match(key)
            if match is not None:
                temp_dict[match.group("index")][match.group("type")] = value[0]

        # 1: {key:X, value:Y}
        # to
        # X: Y
        # All of this, just to take into account when people provide invalid stuff.
        result = {}
        for item in temp_dict.values():
            if "key" in item and "value" in item:
                result[item["key"]] = item["value"]

        self.backend.set_sms_attributes(result)

        return EmptyResult()

    def get_sms_attributes(self) -> ActionResult:
        filter_list = set()
        for key, value in self.querystring.items():
            if key.startswith("attributes.member.1"):
                filter_list.add(value[0])
        attributes = self.backend.get_sms_attributes(filter_list)
        result = {"attributes": attributes}
        return ActionResult(result)

    def check_if_phone_number_is_opted_out(
        self,
    ) -> ActionResult:
        number = self._get_param("phoneNumber")
        if self.OPT_OUT_PHONE_NUMBER_REGEX.match(number) is None:
            raise SNSInvalidParameter(
                "Invalid parameter: PhoneNumber Reason: input incorrectly formatted"
            )

        x = self.backend.check_if_phone_number_is_opted_out(number)
        result = {"isOptedOut": x}
        return ActionResult(result)

    def list_phone_numbers_opted_out(self) -> ActionResult:
        numbers = self.backend.list_phone_numbers_opted_out()
        result = {"phoneNumbers": numbers}
        return ActionResult(result)

    def opt_in_phone_number(self) -> ActionResult:
        number = self._get_param("phoneNumber")
        self.backend.opt_in_phone_number(number)
        return EmptyResult()

    def add_permission(self) -> ActionResult:
        topic_arn = self._get_param("TopicArn")
        label = self._get_param("Label")
        aws_account_ids = self._get_param("AWSAccountId", [])
        action_names = self._get_param("ActionName", [])
        self.backend.add_permission(
            region_name=self.region,
            topic_arn=topic_arn,
            label=label,
            aws_account_ids=aws_account_ids,
            action_names=action_names,
        )
        return EmptyResult()

    def remove_permission(self) -> ActionResult:
        topic_arn = self._get_param("TopicArn")
        label = self._get_param("Label")
        self.backend.remove_permission(topic_arn, label)
        return EmptyResult()

    def confirm_subscription(self) -> ActionResult:
        arn = self._get_param("TopicArn")

        if arn not in self.backend.topics:
            raise SNSNotFoundError("Topic does not exist")

        # Once Tokens are stored by the `subscribe` endpoint and distributed
        # to the client somehow, then we can check validity of tokens
        # presented to this method. The following code works, all thats
        # needed is to perform a token check and assign that value to the
        # `already_subscribed` variable.
        #
        # token = self._get_param('Token')
        # auth = self._get_param('AuthenticateOnUnsubscribe')
        # if already_subscribed:
        #     error_response = self._error(
        #         code='AuthorizationError',
        #         message='Subscription already confirmed'
        #     )
        #     return error_response, dict(status=400)

        result = {"SubscriptionArn": f"{arn}:68762e72-e9b1-410a-8b3b-903da69ee1d5"}
        return ActionResult(result)

    def list_tags_for_resource(self) -> ActionResult:
        arn = self._get_param("ResourceArn")
        tags = self.backend.list_tags_for_resource(arn)
        result = {"Tags": tags}
        return ActionResult(result)

    def tag_resource(self) -> ActionResult:
        arn = self._get_param("ResourceArn")
        tags = self._get_tags()
        self.backend.tag_resource(arn, tags)
        return EmptyResult()

    def untag_resource(self) -> ActionResult:
        arn = self._get_param("ResourceArn")
        tag_keys = self._get_param("TagKeys", [])
        self.backend.untag_resource(arn, tag_keys)
        return EmptyResult()

    @staticmethod
    def serve_pem(request: Any, full_url: str, headers: Any) -> TYPE_RESPONSE:
        sns = SNSResponse()
        sns.automated_parameter_parsing = False
        sns.setup_class(request, full_url, headers)
        key_name = full_url.split("/")[-1]
        key = sns.backend._message_public_keys[key_name]
        return 200, {}, key



from moto.core.responses import ActionResult, EmptyResult

from ..exceptions import AuthFailureRestricted, InvalidRequest
from ._base_response import EC2BaseResponse


class AmisResponse(EC2BaseResponse):
    def create_image(self) -> ActionResult:
        name = self._get_param("Name")
        description = self._get_param("Description", "")
        instance_id = self._get_param("InstanceId")
        tag_specifications = self._get_param("TagSpecifications", [])

        self.error_on_dryrun()

        image = self.ec2_backend.create_image(
            instance_id,
            name,
            description,
            tag_specifications=tag_specifications,
        )
        result = {"ImageId": image.id}
        return ActionResult(result)

    def copy_image(self) -> ActionResult:
        source_image_id = self._get_param("SourceImageId")
        source_region = self._get_param("SourceRegion")
        name = self._get_param("Name")
        description = self._get_param("Description")

        self.error_on_dryrun()

        image = self.ec2_backend.copy_image(
            source_image_id, source_region, name, description
        )
        result = {"ImageId": image.id}
        return ActionResult(result)

    def deregister_image(self) -> ActionResult:
        ami_id = self._get_param("ImageId")

        self.error_on_dryrun()

        self.ec2_backend.deregister_image(ami_id)
        result = {"Return": True}
        return ActionResult(result)

    def describe_images(self) -> ActionResult:
        self.error_on_dryrun()
        ami_ids = self._get_param("ImageIds", [])
        filters = self._filters_from_querystring()
        owners = self._get_param("Owners", [])
        exec_users = self._get_param("ExecutableUsers", [])
        images = self.ec2_backend.describe_images(
            ami_ids=ami_ids, filters=filters, exec_users=exec_users, owners=owners
        )
        result = {"Images": images}
        return ActionResult(result)

    def describe_image_attribute(self) -> ActionResult:
        ami_id = self._get_param("ImageId")
        attribute_name = self._get_param("Attribute")

        # only valid attributes as per
        # https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/ec2/client/describe_image_attribute.html
        valid_attributes_list = {
            "description": "description",
            "kernel": "kernel_id",
            "ramdisk": "ramdisk",
            "launchPermission": "launch_permissions",
            "productCodes": "product_codes",
            "blockDeviceMapping": "bdm",
            "sriovNetSupport": "sriov",
            "bootMode": "boot_mode",
            "tpmSupport": "tmp",
            "uefiData": "uefi",
            "lastLaunchedTime": "lld",
            "imdsSupport": "imds",
        }
        if attribute_name not in valid_attributes_list:
            raise InvalidRequest
        elif attribute_name == "blockDeviceMapping":
            # replicate real aws behaviour and throw and error
            # https://github.com/aws/aws-cli/issues/1083
            raise AuthFailureRestricted

        attribute_value = None
        launch_permissions = None
        if attribute_name == "launchPermission":
            launch_permissions = self.ec2_backend.describe_image_attribute(
                ami_id, valid_attributes_list[attribute_name]
            )
        else:
            attribute_value = self.ec2_backend.describe_image_attribute(
                ami_id, valid_attributes_list[attribute_name]
            )

        result = {"ImageId": ami_id}
        if attribute_name == "productCodes":
            result["ProductCodes"] = [
                {"ProductCodeId": code, "ProductCodeType": "marketplace"}
                for code in attribute_value  # type: ignore[union-attr]
            ]
        elif attribute_name == "launchPermission":
            result["LaunchPermissions"] = launch_permissions
        else:
            result[attribute_name] = {"Value": attribute_value}

        return ActionResult(result)

    def modify_image_attribute(self) -> ActionResult:
        ami_id = self._get_param("ImageId")
        launch_permissions_to_add = self._get_param("LaunchPermission", {}).get(
            "Add", []
        )
        launch_permissions_to_remove = self._get_param("LaunchPermission", {}).get(
            "Remove", []
        )
        # If only one OperationType is added, the other attributes are submitted as different variables
        operation_type = self._get_param("OperationType")
        if operation_type in ["add", "remove"]:
            group = self._get_param("UserGroups", [])
            lp = (
                launch_permissions_to_add
                if operation_type == "add"
                else launch_permissions_to_remove
            )
            if group:
                lp.append({"Group": group[0]})

            for user_id in self._get_param("UserIds", []):
                lp.append({"UserId": user_id})

            org_arn = self._get_param("OrganizationArns", [])
            if org_arn:
                lp.append({"OrganizationArn": org_arn[0]})

            ou_arn = self._get_param("OrganizationalUnitArns", [])
            if ou_arn:
                lp.append({"OrganizationalUnitArn": ou_arn[0]})

        self.error_on_dryrun()

        self.ec2_backend.modify_image_attribute(
            ami_id=ami_id,
            launch_permissions_to_add=launch_permissions_to_add,
            launch_permissions_to_remove=launch_permissions_to_remove,
        )
        return EmptyResult()

    def register_image(self) -> ActionResult:
        name = self._get_param("Name")
        description = self._get_param("Description", "")

        self.error_on_dryrun()

        image = self.ec2_backend.register_image(name, description)
        result = {"ImageId": image.id}
        return ActionResult(result)

    def reset_image_attribute(self) -> str:
        self.error_on_dryrun()

        raise NotImplementedError("AMIs.reset_image_attribute is not yet implemented")



import json
import re
from typing import Any, Optional

import yaml
from yaml.parser import ParserError
from yaml.scanner import ScannerError

from moto.core.common_models import CloudFormationModel
from moto.core.responses import ActionResult, BaseResponse, EmptyResult
from moto.s3.exceptions import S3ClientError

from .exceptions import MissingParameterError, ValidationError
from .models import (
    Change,
    CloudFormationBackend,
    Stack,
    StackSet,
    cloudformation_backends,
)
from .utils import get_stack_from_s3_url, yaml_tag_constructor


class StackResourceDTO:
    def __init__(self, stack: Stack, resource: type[CloudFormationModel]) -> None:
        self.stack_id = stack.stack_id
        self.stack_name = stack.name
        # FIXME: None of these attributes are part of CloudFormationModel interface...
        self.logical_resource_id = getattr(resource, "logical_resource_id", None)
        self.physical_resource_id = getattr(resource, "physical_resource_id", None)
        self.resource_type = getattr(resource, "cf_resource_type", None)
        self.stack_status = stack.status
        self.creation_time = stack.creation_time
        self.timestamp = "2010-07-27T22:27:28Z"  # Hardcoded in original XML template.
        self.resource_status = stack.status


class StackSetOperationDTO:
    def __init__(self, operation: dict[str, Any], stack_set: StackSet) -> None:
        self.execution_role_name = stack_set.execution_role_name
        self.administration_role_arn = stack_set.administration_role_arn
        self.stack_set_id = stack_set.id
        self.creation_timestamp = operation["CreationTimestamp"]
        self.operation_id = operation["OperationId"]
        self.action = operation["Action"]
        self.end_timestamp = operation.get("EndTimestamp", None)
        self.status = operation["Status"]


class ChangeDTO:
    def __init__(self, change: Change) -> None:
        self.type = "Resource"
        self.resource_change = change


def get_template_summary_response_from_template(template_body: str) -> dict[str, Any]:
    def get_resource_types(template_dict: dict[str, Any]) -> list[Any]:
        resources = {}
        for key, value in template_dict.items():
            if key == "Resources":
                resources = value

        resource_types = []
        for value in resources.values():
            resource_types.append(value["Type"])
        return resource_types

    yaml.add_multi_constructor("", yaml_tag_constructor)

    try:
        template_dict = yaml.load(template_body, Loader=yaml.Loader)
    except (ParserError, ScannerError):
        template_dict = json.loads(template_body)

    resources_types = get_resource_types(template_dict)
    template_dict["ResourceTypes"] = resources_types
    template_dict["Version"] = template_dict["AWSTemplateFormatVersion"]
    parameters = []
    for key, value in template_dict.get("Parameters", {}).items():
        parameter = {
            "ParameterKey": key,
            "Description": value.get("Description", ""),
            "DefaultValue": value.get("Default", None),
            "NoEcho": value.get("NoEcho", False),
            "ParameterType": value.get("Type", "String"),
            "ParameterConstraints": {},
        }
        if "AllowedValues" in value:
            parameter["ParameterConstraints"]["AllowedValues"] = value["AllowedValues"]
        parameters.append(parameter)
    template_dict["Parameters"] = parameters
    return template_dict


def transform_dict(
    data: dict[str, str], key_for_key: str = "Key", key_for_value: str = "Value"
) -> list[dict[str, str]]:
    transformed = [
        {key_for_key: key, key_for_value: value} for key, value in data.items()
    ]
    return transformed


def transform_parameters(data: dict[str, str]) -> list[dict[str, str]]:
    return transform_dict(
        data, key_for_key="ParameterKey", key_for_value="ParameterValue"
    )


class CloudFormationResponse(BaseResponse):
    RESPONSE_KEY_PATH_TO_TRANSFORMER = {
        "DescribeChangeSetOutput.Changes": lambda x: [ChangeDTO(c) for c in x],
        "DescribeChangeSetOutput.Parameters": transform_parameters,
        "DescribeStackSetOutput.StackSet.Parameters": transform_parameters,
        "DescribeStackSetOutput.StackSet.Tags": transform_dict,
        "DescribeStacksOutput.Stacks.Stack.Tags": transform_dict,
    }

    def __init__(self) -> None:
        super().__init__(service_name="cloudformation")
        self.automated_parameter_parsing = True

    @property
    def cloudformation_backend(self) -> CloudFormationBackend:
        return cloudformation_backends[self.current_account][self.region]

    @classmethod
    def cfnresponse(cls, *args: Any, **kwargs: Any) -> Any:  # type: ignore[misc]
        request, full_url, headers = args
        full_url += "&Action=ProcessCfnResponse"
        cf = CloudFormationResponse()
        cf.automated_parameter_parsing = False
        return cf._dispatch(request=request, full_url=full_url, headers=headers)

    def _get_stack_from_s3_url(self, template_url: str) -> str:
        return get_stack_from_s3_url(
            template_url, account_id=self.current_account, partition=self.partition
        )

    def _get_params_from_list(
        self, parameters_list: list[dict[str, Any]]
    ) -> dict[str, Any]:
        # Hack dict-comprehension
        return {
            parameter["ParameterKey"]: parameter["ParameterValue"]
            for parameter in parameters_list
        }

    def _get_param_values(
        self, parameters_list: list[dict[str, str]], existing_params: dict[str, str]
    ) -> dict[str, Any]:
        result = {}
        for parameter in parameters_list:
            if set(parameter.keys()) >= {"ParameterKey", "ParameterValue"}:
                result[parameter["ParameterKey"]] = parameter["ParameterValue"]
            elif (
                set(parameter.keys()) >= {"ParameterKey", "UsePreviousValue"}
                and parameter["ParameterKey"] in existing_params
            ):
                result[parameter["ParameterKey"]] = existing_params[
                    parameter["ParameterKey"]
                ]
            else:
                raise MissingParameterError(parameter["ParameterKey"])
        return result

    def process_cfn_response(self) -> tuple[int, dict[str, int], str]:
        status = self._get_param("Status")
        if status == "SUCCESS":
            stack_id = self._get_param("StackId")
            logical_resource_id = self._get_param("LogicalResourceId")
            outputs = self._get_param("Data")
            stack = self.cloudformation_backend.get_stack(stack_id)
            custom_resource = stack.get_custom_resource(logical_resource_id)
            custom_resource.set_data(outputs)
            stack.verify_readiness()

        return 200, {"status": 200}, json.dumps("{}")

    def create_stack(self) -> ActionResult:
        stack_name = self._get_param("StackName")
        stack_body = self._get_param("TemplateBody")
        template_url = self._get_param("TemplateURL")
        role_arn = self._get_param("RoleARN")
        enable_termination_protection = self._get_param("EnableTerminationProtection")
        timeout_in_mins = self._get_param("TimeoutInMinutes")
        stack_policy_body = self._get_param("StackPolicyBody")
        parameters_list = self._get_param("Parameters", [])
        tags = {item["Key"]: item["Value"] for item in self._get_param("Tags", [])}

        parameters = self._get_params_from_list(parameters_list)

        if template_url:
            stack_body = self._get_stack_from_s3_url(template_url)
        stack_notification_arns = self._get_param("NotificationARNs", [])

        stack = self.cloudformation_backend.create_stack(
            name=stack_name,
            template=stack_body,
            parameters=parameters,
            notification_arns=stack_notification_arns,
            tags=tags,
            role_arn=role_arn,
            enable_termination_protection=enable_termination_protection,
            timeout_in_mins=timeout_in_mins,
            stack_policy_body=stack_policy_body,
        )
        result = {"StackId": stack.stack_id}
        return ActionResult(result)

    def validate_template_and_stack_body(self) -> None:
        if (
            self._get_param("TemplateBody") or self._get_param("TemplateURL")
        ) and self._get_bool_param("UsePreviousTemplate", False):
            raise ValidationError(
                message="An error occurred (ValidationError) when calling the CreateChangeSet operation: You cannot specify both usePreviousTemplate and Template Body/Template URL."
            )
        elif (
            not self._get_param("TemplateBody")
            and not self._get_param("TemplateURL")
            and not self._get_bool_param("UsePreviousTemplate", False)
        ):
            raise ValidationError(
                message="An error occurred (ValidationError) when calling the CreateChangeSet operation: Either Template URL or Template Body must be specified."
            )

    def create_change_set(self) -> ActionResult:
        stack_name = self._get_param("StackName")
        change_set_name = self._get_param("ChangeSetName")
        stack_body = self._get_param("TemplateBody")
        template_url = self._get_param("TemplateURL")
        update_or_create = self._get_param("ChangeSetType", "CREATE")
        use_previous_template = self._get_bool_param("UsePreviousTemplate", False)
        if update_or_create == "UPDATE":
            stack = self.cloudformation_backend.get_stack(stack_name)
            self.validate_template_and_stack_body()
            if use_previous_template:
                stack_body = stack.template
        description = self._get_param("Description")
        role_arn = self._get_param("RoleARN")
        parameters_list = self._get_param("Parameters", [])
        tags = {item["Key"]: item["Value"] for item in self._get_param("Tags", [])}
        parameters = {
            param["ParameterKey"]: (
                stack.stack_parameters[param["ParameterKey"]]
                if param.get("UsePreviousValue", False)
                else param["ParameterValue"]
            )
            for param in parameters_list
        }
        if update_or_create == "UPDATE":
            self._validate_different_update(parameters_list, stack_body, stack)

        if template_url:
            stack_body = self._get_stack_from_s3_url(template_url)
        stack_notification_arns = self._get_param("NotificationARNs", [])
        change_set_id, stack_id = self.cloudformation_backend.create_change_set(
            stack_name=stack_name,
            change_set_name=change_set_name,
            template=stack_body,
            parameters=parameters,
            description=description,
            notification_arns=stack_notification_arns,
            tags=tags,
            role_arn=role_arn,
            change_set_type=update_or_create,
        )
        result = {"Id": change_set_id, "StackId": stack_id}
        return ActionResult(result)

    def delete_change_set(self) -> ActionResult:
        change_set_name = self._get_param("ChangeSetName")

        self.cloudformation_backend.delete_change_set(change_set_name=change_set_name)
        return EmptyResult()

    def describe_change_set(self) -> ActionResult:
        change_set_name = self._get_param("ChangeSetName")
        change_set = self.cloudformation_backend.describe_change_set(
            change_set_name=change_set_name
        )
        return ActionResult(change_set)

    def execute_change_set(self) -> ActionResult:
        stack_name = self._get_param("StackName")
        change_set_name = self._get_param("ChangeSetName")
        self.cloudformation_backend.execute_change_set(
            stack_name=stack_name, change_set_name=change_set_name
        )
        return EmptyResult()

    def describe_stacks(self) -> ActionResult:
        stack_name_or_id = self._get_param("StackName")
        token = self._get_param("NextToken")
        stacks = self.cloudformation_backend.describe_stacks(stack_name_or_id)
        stack_ids = [stack.stack_id for stack in stacks]
        if token:
            start = stack_ids.index(token) + 1
        else:
            start = 0
        max_results = 50  # using this to mske testing of paginated stacks more convenient than default 1 MB
        stacks_resp = stacks[start : start + max_results]
        next_token = None
        if len(stacks) > (start + max_results):
            next_token = stacks_resp[-1].stack_id
        result = {"Stacks": stacks_resp, "NextToken": next_token}
        return ActionResult(result)

    def describe_stack_resource(self) -> ActionResult:
        stack_name = self._get_param("StackName")
        logical_resource_id = self._get_param("LogicalResourceId")
        stack, resource = self.cloudformation_backend.describe_stack_resource(
            stack_name, logical_resource_id
        )
        result = {"StackResourceDetail": StackResourceDTO(stack, resource)}
        return ActionResult(result)

    def describe_stack_resources(self) -> ActionResult:
        stack_name = self._get_param("StackName")
        logical_resource_id = self._get_param("LogicalResourceId")
        stack, resources = self.cloudformation_backend.describe_stack_resources(
            stack_name, logical_resource_id
        )

        result = {
            "StackResources": [
                StackResourceDTO(stack, resource) for resource in resources
            ]
        }
        return ActionResult(result)

    def describe_stack_events(self) -> ActionResult:
        stack_name = self._get_param("StackName")
        events = self.cloudformation_backend.describe_stack_events(stack_name)

        result = {"StackEvents": events[::-1]}
        return ActionResult(result)

    def list_change_sets(self) -> ActionResult:
        change_sets = self.cloudformation_backend.list_change_sets()
        result = {"Summaries": change_sets}
        return ActionResult(result)

    def list_stacks(self) -> ActionResult:
        status_filter = self._get_param("StackStatusFilter", [])
        stacks = self.cloudformation_backend.list_stacks(status_filter)
        result = {"StackSummaries": stacks}
        return ActionResult(result)

    def list_stack_resources(self) -> ActionResult:
        stack_name_or_id = self._get_param("StackName")
        resources = self.cloudformation_backend.list_stack_resources(stack_name_or_id)
        # FIXME: This was migrated from the original XML template, including hardcoded values.
        result = {
            "StackResourceSummaries": [
                {
                    "ResourceStatus": "CREATE_COMPLETE",
                    "LogicalResourceId": getattr(resource, "logical_resource_id", ""),
                    "LastUpdateTimestamp": "2011-06-21T20:15:58Z",
                    "PhysicalResourceId": getattr(resource, "physical_resource_id", ""),
                    "ResourceType": getattr(resource, "cf_resource_type", ""),
                }
                for resource in resources
            ]
        }
        return ActionResult(result)

    def get_template(self) -> ActionResult:
        name_or_stack_id = self._get_param("StackName")
        stack_template = self.cloudformation_backend.get_template(name_or_stack_id)
        result = {"TemplateBody": stack_template}
        return ActionResult(result)

    def get_template_summary(self) -> ActionResult:
        stack_name = self._get_param("StackName")
        template_url = self._get_param("TemplateURL")
        stack_body = self._get_param("TemplateBody")

        if stack_name:
            stack = self.cloudformation_backend.get_stack(stack_name)
            if stack.status == "REVIEW_IN_PROGRESS":
                raise ValidationError(
                    message="GetTemplateSummary cannot be called on REVIEW_IN_PROGRESS stacks."
                )
            stack_body = stack.template
        elif template_url:
            stack_body = self._get_stack_from_s3_url(template_url)

        template_summary = get_template_summary_response_from_template(stack_body)
        return ActionResult(template_summary)

    def _validate_different_update(
        self,
        incoming_params: Optional[list[dict[str, Any]]],
        stack_body: str,
        old_stack: Stack,
    ) -> None:
        if incoming_params and stack_body:
            new_params = self._get_param_values(
                incoming_params, old_stack.stack_parameters
            )
            if (
                old_stack.template == stack_body
                and old_stack.stack_parameters == new_params
            ):
                raise ValidationError(
                    old_stack.name, message="No updates are to be performed."
                )

    def _validate_status(self, stack: Stack) -> None:
        if stack.status == "ROLLBACK_COMPLETE":
            raise ValidationError(
                stack.stack_id,
                message=f"Stack:{stack.stack_id} is in ROLLBACK_COMPLETE state and can not be updated.",
            )

    def update_stack(self) -> ActionResult:
        stack_name = self._get_param("StackName")
        role_arn = self._get_param("RoleARN")
        template_url = self._get_param("TemplateURL")
        stack_body = self._get_param("TemplateBody")
        stack = self.cloudformation_backend.get_stack(stack_name)
        if self._get_bool_param("UsePreviousTemplate", False):
            stack_body = stack.template
        elif not stack_body and template_url:
            stack_body = self._get_stack_from_s3_url(template_url)

        incoming_params = self._get_param("Parameters", [])
        for param in incoming_params:
            if param.get("UsePreviousValue") and param.get("ParameterValue"):
                raise ValidationError(
                    message=f"Invalid input for parameter key {param['ParameterKey']}. Cannot specify usePreviousValue as true and non empty value for a parameter"
                )
        # boto3 is supposed to let you clear the tags by passing an empty value, but the request body doesn't
        # end up containing anything we can use to differentiate between passing an empty value versus not
        # passing anything. so until that changes, moto won't be able to clear tags, only update them.
        tags: Optional[dict[str, str]] = {
            item["Key"]: item["Value"] for item in self._get_param("Tags", [])
        }
        # so that if we don't pass the parameter, we don't clear all the tags accidentally
        if not tags:
            tags = None

        stack = self.cloudformation_backend.get_stack(stack_name)
        self._validate_different_update(incoming_params, stack_body, stack)
        self._validate_status(stack)

        stack = self.cloudformation_backend.update_stack(
            name=stack_name,
            template=stack_body,
            role_arn=role_arn,
            parameters=incoming_params,
            tags=tags,
        )
        result = {"StackId": stack.stack_id}
        return ActionResult(result)

    def delete_stack(self) -> ActionResult:
        name_or_stack_id = self._get_param("StackName")
        self.cloudformation_backend.delete_stack(name_or_stack_id)
        return EmptyResult()

    def list_exports(self) -> ActionResult:
        token = self._get_param("NextToken")
        exports, next_token = self.cloudformation_backend.list_exports(tokenstr=token)
        result = {"Exports": exports, "NextToken": next_token}
        return ActionResult(result)

    def validate_template(self) -> ActionResult:
        template_body = self._get_param("TemplateBody")
        template_url = self._get_param("TemplateURL")
        if template_url:
            template_body = self._get_stack_from_s3_url(template_url)

        cfn_lint = self.cloudformation_backend.validate_template(template_body)
        if cfn_lint:
            raise ValidationError(cfn_lint[0].message)
        description = ""
        try:
            description = json.loads(template_body)["Description"]
        except (ValueError, KeyError):
            pass
        try:
            yaml.add_multi_constructor("", yaml_tag_constructor)
            description = yaml.load(template_body, Loader=yaml.Loader)["Description"]
        except (ParserError, ScannerError, KeyError):
            pass
        result = {
            "Description": description,
            "Parameters": [],
            "Capabilities": [],
            "DeclaredTransforms": [],
        }
        return ActionResult(result)

    def create_stack_set(self) -> ActionResult:
        stackset_name = self._get_param("StackSetName")
        if not re.match(r"^[a-zA-Z][-a-zA-Z0-9]*$", stackset_name):
            raise ValidationError(
                message=f"1 validation error detected: Value '{stackset_name}' at 'stackSetName' failed to satisfy constraint: Member must satisfy regular expression pattern: [a-zA-Z][-a-zA-Z0-9]*"
            )
        stack_body = self._get_param("TemplateBody")
        template_url = self._get_param("TemplateURL")
        permission_model = self._get_param("PermissionModel")
        parameters_list = self._get_param("Parameters", [])
        admin_role = self._get_param("AdministrationRoleARN")
        exec_role = self._get_param("ExecutionRoleName")
        description = self._get_param("Description")
        tags = {item["Key"]: item["Value"] for item in self._get_param("Tags", [])}

        # Copy-Pasta - Hack dict-comprehension
        parameters = {
            parameter["ParameterKey"]: parameter["ParameterValue"]
            for parameter in parameters_list
        }
        if template_url:
            stack_body = self._get_stack_from_s3_url(template_url)

        stackset = self.cloudformation_backend.create_stack_set(
            name=stackset_name,
            template=stack_body,
            parameters=parameters,
            tags=tags,
            permission_model=permission_model,
            admin_role=admin_role,
            exec_role=exec_role,
            description=description,
        )
        result = {"StackSetId": stackset.id}
        return ActionResult(result)

    def create_stack_instances(self) -> ActionResult:
        stackset_name = self._get_param("StackSetName")
        accounts = self._get_param("Accounts", [])
        regions = self._get_param("Regions", [])
        parameters = self._get_param("ParameterOverrides", [])
        deployment_targets = self._get_params().get("DeploymentTargets")
        if deployment_targets and "OrganizationalUnitIds" in deployment_targets:
            for ou_id in deployment_targets.get("OrganizationalUnitIds", []):
                if not re.match(
                    r"^(ou-[a-z0-9]{4,32}-[a-z0-9]{8,32}|r-[a-z0-9]{4,32})$", ou_id
                ):
                    raise ValidationError(
                        message=f"1 validation error detected: Value '[{ou_id}]' at 'deploymentTargets.organizationalUnitIds' failed to satisfy constraint: Member must satisfy constraint: [Member must have length less than or equal to 68, Member must have length greater than or equal to 6, Member must satisfy regular expression pattern: ^(ou-[a-z0-9]{{4,32}}-[a-z0-9]{{8,32}}|r-[a-z0-9]{{4,32}})$]"
                    )

        operation_id = self.cloudformation_backend.create_stack_instances(
            stackset_name,
            accounts,
            regions,
            parameters,
            deployment_targets=deployment_targets,
        )
        result = {"OperationId": operation_id}
        return ActionResult(result)

    def delete_stack_set(self) -> ActionResult:
        stackset_name = self._get_param("StackSetName")
        self.cloudformation_backend.delete_stack_set(stackset_name)
        return EmptyResult()

    def delete_stack_instances(self) -> ActionResult:
        stackset_name = self._get_param("StackSetName")
        accounts = self._get_param("Accounts", [])
        regions = self._get_param("Regions", [])
        operation = self.cloudformation_backend.delete_stack_instances(
            stackset_name, accounts, regions
        )

        result = {"OperationId": operation.operations[-1]["OperationId"]}
        return ActionResult(result)

    def describe_stack_set(self) -> ActionResult:
        stackset_name = self._get_param("StackSetName")
        stackset = self.cloudformation_backend.describe_stack_set(stackset_name)
        result = {"StackSet": stackset}
        return ActionResult(result)

    def describe_stack_instance(self) -> ActionResult:
        stackset_name = self._get_param("StackSetName")
        account = self._get_param("StackInstanceAccount")
        region = self._get_param("StackInstanceRegion")

        instance = self.cloudformation_backend.describe_stack_instance(
            stackset_name, account, region
        )
        result = {"StackInstance": instance}
        return ActionResult(result)

    def list_stack_sets(self) -> ActionResult:
        stacksets = self.cloudformation_backend.list_stack_sets()
        result = {"Summaries": stacksets}
        return ActionResult(result)

    def list_stack_instances(self) -> ActionResult:
        stackset_name = self._get_param("StackSetName")
        instances = self.cloudformation_backend.list_stack_instances(stackset_name)
        result = {"Summaries": instances}
        return ActionResult(result)

    def list_stack_set_operations(self) -> ActionResult:
        stackset_name = self._get_param("StackSetName")
        operations = self.cloudformation_backend.list_stack_set_operations(
            stackset_name
        )
        result = {"Summaries": operations}
        return ActionResult(result)

    def stop_stack_set_operation(self) -> ActionResult:
        stackset_name = self._get_param("StackSetName")
        operation_id = self._get_param("OperationId")
        self.cloudformation_backend.stop_stack_set_operation(
            stackset_name, operation_id
        )
        return EmptyResult()

    def describe_stack_set_operation(self) -> ActionResult:
        stackset_name = self._get_param("StackSetName")
        operation_id = self._get_param("OperationId")
        stackset, operation = self.cloudformation_backend.describe_stack_set_operation(
            stackset_name, operation_id
        )
        result = {"StackSetOperation": StackSetOperationDTO(operation, stackset)}
        return ActionResult(result)

    def list_stack_set_operation_results(self) -> ActionResult:
        stackset_name = self._get_param("StackSetName")
        operation_id = self._get_param("OperationId")
        operation = self.cloudformation_backend.list_stack_set_operation_results(
            stackset_name, operation_id
        )
        # FIXME: Hardcoded response values come from original XML template.
        result = {
            "Summaries": [
                {
                    "AccountGateResult": {
                        "StatusReason": f"Function not found: arn:aws:lambda:us-west-2:{account}:function:AWSCloudFormationStackSetAccountGate",
                        "Status": "SKIPPED",
                    },
                    "Region": region,
                    "Account": account,
                    "Status": operation["Status"],
                }
                for instance in operation["Instances"]
                for account, region in instance.items()
            ]
        }
        return ActionResult(result)

    def update_stack_set(self) -> ActionResult:
        stackset_name = self._get_param("StackSetName")
        operation_id = self._get_param("OperationId")
        description = self._get_param("Description")
        execution_role = self._get_param("ExecutionRoleName")
        admin_role = self._get_param("AdministrationRoleARN")
        accounts = self._get_param("Accounts", [])
        regions = self._get_param("Regions", [])
        template_body = self._get_param("TemplateBody")
        template_url = self._get_param("TemplateURL")
        if template_url:
            template_body = self._get_stack_from_s3_url(template_url)
        tags = {item["Key"]: item["Value"] for item in self._get_param("Tags", [])}
        parameters_list = self._get_param("Parameters", [])

        operation = self.cloudformation_backend.update_stack_set(
            stackset_name=stackset_name,
            template=template_body,
            description=description,
            parameters=parameters_list,
            tags=tags,
            admin_role=admin_role,
            execution_role=execution_role,
            accounts=accounts,
            regions=regions,
            operation_id=operation_id,
        )

        result = {"OperationId": operation["OperationId"]}
        return ActionResult(result)

    def update_stack_instances(self) -> ActionResult:
        stackset_name = self._get_param("StackSetName")
        accounts = self._get_param("Accounts", [])
        regions = self._get_param("Regions", [])
        parameters = self._get_param("ParameterOverrides", [])
        operation = self.cloudformation_backend.update_stack_instances(
            stackset_name, accounts, regions, parameters
        )
        result = {"OperationId": operation["OperationId"]}
        return ActionResult(result)

    def get_stack_policy(self) -> ActionResult:
        stack_name = self._get_param("StackName")
        policy = self.cloudformation_backend.get_stack_policy(stack_name)
        result = {"StackPolicyBody": policy if policy else None}
        return ActionResult(result)

    def set_stack_policy(self) -> ActionResult:
        stack_name = self._get_param("StackName")
        policy_url = self._get_param("StackPolicyURL")
        policy_body = self._get_param("StackPolicyBody")
        if policy_body and policy_url:
            raise ValidationError(
                message="You cannot specify both StackPolicyURL and StackPolicyBody"
            )
        if policy_url:
            try:
                policy_body = self._get_stack_from_s3_url(policy_url)
            except S3ClientError as s3_e:
                raise ValidationError(
                    message=f"S3 error: Access Denied: {s3_e.error_type}"
                )
        self.cloudformation_backend.set_stack_policy(
            stack_name, policy_body=policy_body
        )
        return EmptyResult()



from collections import defaultdict

"""
Storage of all instances that extend BaseModel
This allows us to easily expose all internal state using the MotoServer dashboard
"""
model_data: dict[str, dict[str, object]] = defaultdict(dict)


def reset_model_data() -> None:
    # Remove all references to the models stored
    for models in model_data.values():
        for model in models.values():
            model.instances_tracked = []  # type: ignore[attr-defined]



from enum import Enum

VALID_AUTH_MODE_KEYS = ["Type", "Passwords"]
VALID_ENGINE_TYPES = ["redis", "valkey"]


class AuthenticationTypes(str, Enum):
    NOPASSWORD = "no-password-required"
    PASSWORD = "password"
    IAM = "iam"



from typing import Any

from moto.core.exceptions import JsonRESTError


class MacieException(JsonRESTError):
    pass


class ValidationException(MacieException):
    code = 400

    def __init__(self, message: str):
        super().__init__("ValidationException", message)

    def get_headers(self, *args: Any, **kwargs: Any) -> list[tuple[str, str]]:
        return [("x-amzn-ErrorType", "ValidationException")]


class AccessDeniedException(MacieException):
    code = 403

    def __init__(self, message: str):
        super().__init__("AccessDeniedException", message)

    def get_headers(self, *args: Any, **kwargs: Any) -> list[tuple[str, str]]:
        return [("x-amzn-ErrorType", "AccessDeniedException")]


class ResourceNotFoundException(MacieException):
    code = 404

    def __init__(self, message: str):
        super().__init__("ResourceNotFoundException", message)

    def get_headers(self, *args: Any, **kwargs: Any) -> list[tuple[str, str]]:
        return [("x-amzn-ErrorType", "ResourceNotFoundException")]



from collections.abc import Iterator
from typing import Union


def string_equals_operation(
    expected_value: Union[list[str], str], actual_value: str
) -> bool:
    if isinstance(expected_value, str):
        return actual_value == expected_value

    if isinstance(expected_value, list):
        return actual_value in expected_value

    return False


CONDITION_OPERATIONS = {"StringEquals": string_equals_operation}


class TrustCondition:
    """Single condition object"""

    def __init__(
        self,
        condition: str,
        expected_value: list[Union[list[str], str]],
        expected_value_source: list[str],
    ) -> None:
        self._condition = condition

        self._expected_value_source = expected_value_source
        self._expected_value = expected_value

    @property
    def context_keys(self) -> list[str]:
        return self._expected_value_source

    def verify_condition(self, actual_values: list[str]) -> bool:
        verify_action = CONDITION_OPERATIONS.get(self._condition)

        # If the evaluation operation is not supported, ignore condition evaluation
        if not verify_action:
            return True

        for index, actual_value in enumerate(actual_values):
            if not verify_action(self._expected_value[index], actual_value):
                return False

        return True


ConditionData = dict[str, dict[str, Union[str, list[str]]]]


class TrustRelationShipConditions:
    """Trust relationship conditions segment model"""

    def __init__(self, conditions: ConditionData) -> None:
        self._conditions: list[TrustCondition] = []
        for conditions_operation, condition_values in conditions.items():
            expected_value_source: list[str] = list(condition_values.keys())
            expected_value_data: list[Union[str, list[str]]] = list(
                condition_values.values()
            )

            self._conditions.append(
                TrustCondition(
                    conditions_operation, expected_value_data, expected_value_source
                )
            )

    def __iter__(self) -> Iterator[TrustCondition]:
        return self._conditions.__iter__()
